#!/usr/bin/env python
"""
Pipeline for the annotation of transposable elements (TEs) in genomic sequences.
"""

import os
import sys
import getopt
import time
import ConfigParser
import glob
import shutil

if not os.environ.has_key( "REPET_PATH" ):
    print "ERROR: no environment variable REPET_PATH"
    sys.exit(1)
sys.path.append( os.environ["REPET_PATH"] )
import pyRepet.launcher.programLauncher
from pyRepetUnit.longJoinsForTEs.LongJoinsForTEs import LongJoinsForTEs
from pyRepetUnit.commons.coord.AlignUtils import AlignUtils
from pyRepet.util.Stat import Stat
from pyRepetUnit.commons.checker.CheckerUtils import CheckerUtils
from pyRepetUnit.commons.checker.CheckerException import CheckerException
from ConfigParser import NoOptionError
from ConfigParser import NoSectionError
from ConfigParser import MissingSectionHeaderError

#------------------------------------------------------------------------------

def setup_env():

    """
    Setup the required environment for MySQL.
    """

    os.environ["REPET_HOST"] = config.get("repet_env","repet_host")
    os.environ["REPET_USER"] = config.get("repet_env","repet_user")
    os.environ["REPET_PW"] = config.get("repet_env","repet_pw")
    os.environ["REPET_DB"] = config.get("repet_env","repet_db")
    os.environ["REPET_PORT"] = config.get("repet_env","repet_port")

#------------------------------------------------------------------------------

def help():

    """
    Give the list of the command-line options.
    """

    print
    print "usage: ",sys.argv[0],"[ options ]"
    print "options:"
    print "     -h: this help"
    print "     -P: project name (<=15 character only alphanumeric or underscore)"
    print "     -C: configuration file"
    print "     -S: step (1/2 with -a/3 with -c/4 with -s/5/6 with -b/7/8 with -o)"
    print "     -a: program to align sequences (BLR/RM/CEN)"
    print "     -r: use randomized chunks (with \"-S 2\")"
    print "     -c: combine the alignment programs (default=BLR+RM+CEN)"
    print "     -s: program to detect short repeats (TRF/Mreps/RMSSR)"
    print "     -b: blast for step 6 (tblastx/blastx)"
    print "     -o: output format for the annotations (gameXML or GFF3)"
    print "     -v: verbose (0/default=1/2/3)"
    print

#------------------------------------------------------------------------------

def main():

    """
    Pipeline for the annotation of transposable elements (TEs) in genomic sequences.
    """

    global projectName
    projectName = ""
    global configFileName
    configFileName = ""
    step = ""
    align = ""
    useRndChunks = False
    combAlign = "BLR+RM+CEN"
    sat = ""
    typeBlast = ""
    outFormat = ""
    global verbose
    verbose = 1

    try:
        opts,args = getopt.getopt( sys.argv[1:], "hP:C:S:a:rc:s:b:o:v:" )
    except getopt.GetoptError, err:
        print str(err); help(); sys.exit(1)
    for o,a in opts:
        if o == "-h":
            help()
            sys.exit(0)
        elif o == "-P":
            projectName = a
        elif o == "-C":
            configFileName = a
        elif o == "-S":
            step = a
        elif o == "-a":
            align = a
        elif o == "-r":
            useRndChunks = True
        elif o == "-c":
            combAlign = a
        elif o == "-s":
            sat = a
        elif o == "-b":
            typeBlast = a
        elif o == "-o":
            outFormat = a
        elif o == "-v":
            verbose = int(a)

    if projectName == "" or configFileName == "" or step == "":
        print "ERROR: no project provided (-P)"
        help()
        sys.exit(1)
    if configFileName == "":
        print "ERROR: no configuration file provided (-C)"
        help()
        sys.exit(1)
    if step == "":
        print "ERROR: no step provided (-S)"
        help()
        sys.exit(1)
        
    if not CheckerUtils.isMax15Char(projectName):
        print "ERROR: project name must have 15 character max"
        help()
        sys.exit(1)
        
    if not CheckerUtils.isCharAlphanumOrUnderscore(projectName):
        print "ERROR: project name must contain only alphanumeric or underscore character"
        help()
        sys.exit(1)

    global config
    config = ConfigParser.ConfigParser()
    try:
        config.readfp( open(configFileName) )
    except MissingSectionHeaderError:
        print "ERROR: your config file " + configFileName + " must begin with a section name"
        sys.exit(1)

    isExceptionRaised = False
    sectionName = "repet_env"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName)
    except NoSectionError:
        print "ERROR: the section " + sectionName + " must be in your config file : " + configFileName
        sys.exit(1)
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "repet_version") 
    except NoOptionError:
        print "ERROR: the option repet_version must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "repet_host") 
    except NoOptionError:
        print "ERROR: the option repet_host must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "repet_user") 
    except NoOptionError:
        print "ERROR: the option repet_user must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "repet_pw") 
    except NoOptionError:
        print "ERROR: the option repet_pw must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "repet_db") 
    except NoOptionError:
        print "ERROR: the option repet_db must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "repet_port") 
    except NoOptionError:
        print "ERROR: the option repet_port must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    if isExceptionRaised: 
        sys.exit(1)

    setup_env()

    changeLogFileName = os.environ["REPET_PATH"] + "/CHANGELOG" 
    changeLogFileHandler = open(changeLogFileName, "r")
    try:
        CheckerUtils.checkConfigVersion(changeLogFileHandler, config)
    except CheckerException, e:
        print e.msg
        sys.exit(1)
    
    isExceptionRaised = False
    sectionName = "project"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName)
    except NoSectionError:
        print "ERROR: the section " + sectionName + " must be in your config file : " + configFileName
        sys.exit(1)
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "project_name") 
    except NoOptionError:
        print "ERROR: the option project_name must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "project_dir") 
    except NoOptionError:
        print "ERROR: the option project_dir must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    if isExceptionRaised: 
        sys.exit(1)


    if verbose > 0:
        print "\nSTART %s" % (sys.argv[0].split("/")[-1])
        print "version %s" % ( config.get("repet_env","repet_version") )
        now = time.localtime()
        print "local date/time = %s-%s-%s / %s:%s:%s" % ( now[0], now[1], now[2], now[3], now[4], now[5] )
        sys.stdout.flush()

    if config.get( "project", "project_name" ) != projectName:
        print "ERROR: project name different between configuration file ('%s') and command-line option ('%s')" % ( config.get( "project", "project_name" ), projectName )
        sys.exit(1)
    if verbose > 0:
        print "project name = %s" % ( projectName )

    global projectDir
    projectDir = config.get( "project", "project_dir" )
    if projectDir[-1] == "/":
        projectDir = projectDir[:-1]
    if os.path.exists( projectDir ):
        if verbose > 0:
            print "project directory = %s" % ( projectDir ); sys.stdout.flush()
        os.chdir( projectDir )
    else:
        print "ERROR: project directory %s doesn't exist" % ( projectDir )
        sys.exit(1)

    global pL
    pL = pyRepet.launcher.programLauncher.programLauncher()

    #--------------------------------------------------------------------------
            
    if step == "1":            
        prepareInputFilesAndLoadTables()

    elif step == "2":
        if align == "":
            print "ERROR: missing option -a"
            help()
            sys.exit(1)
        if useRndChunks == False:
            alignChunksAndTEs( "", align )
        else:
            alignChunksAndTEs( "_rnd", align )

    elif step == "3":
        filterAndCombineTEsAnnot( combAlign )

    elif step == "4":
        if sat == "":
            print "ERROR: missing option -s"
            help()
            sys.exit(1)
        detectSSRs( sat )

    elif step == "5":
        filterAndCombineSSRsAnnot()

    elif step == "6":
        if typeBlast in [ "tblastx", "blastx" ]:
            alignChunksAndTEsForCoding( typeBlast )
        else:
            print "ERROR: blast '%s' not recognized" % ( typeBlast )
            sys.exit(1)

    elif step == "7":
        pathsProcessing( config )

    elif step == "8":
        if outFormat == "":
            print "ERROR: missing option -o"
            help()
            sys.exit(1)
        exportAnnotations( outFormat )

    if verbose > 0:
        now = time.localtime()
        print "local date/time = %s-%s-%s / %s:%s:%s" % ( now[0], now[1], now[2], now[3], now[4], now[5] )
        print "version %s" % ( config.get("repet_env","repet_version") )
        print "END %s\n" % (sys.argv[0].split("/")[-1])
        sys.stdout.flush()

    return 0

#------------------------------------------------------------------------------

def prepareInputFilesAndLoadTables():

    """
    Step 1:
    * cut the input genomic sequences into chunks and load them in a MySQL table
    * randomize the chunks and load them in a MySQL table
    * load the reference TE library in a MySQL table
    * prepare the reference TE library for Blaster (blastn)
    """
    
    if verbose > 0:
        print "\nbeginning of step 1"
        print "prepare the input files and load them in MySQL tables"
        sys.stdout.flush()
    
    genomeFastaFileName = "%s/%s.fa" % ( projectDir, projectName ) 
    
    if not os.path.exists( genomeFastaFileName ):
        print "ERROR: %s/%s.fa doesn't exist" % ( projectDir, projectName )
        sys.exit(1)
        
    separator = "\n"    
    
    inGenomeFileHandler = open(genomeFastaFileName, "r")
        
    try:
        CheckerUtils.checkHeaders(inGenomeFileHandler)
    except CheckerException, e:
        print "Error in file " + genomeFastaFileName + ". Wrong headers are :"
        print separator.join(e.messages)
        print "Authorized characters are : a-z A-Z 0-9 - . : _ | =\n"
        inGenomeFileHandler.close()
        sys.exit(1)
            
    inGenomeFileHandler.close()
            

    refTEFileName = "%s/%s_refTEs.fa" % ( projectDir, projectName )
    if not os.path.exists( refTEFileName ):
        print "ERROR: %s/%s_refTEs.fa doesn't exist" % ( projectDir, projectName )
        sys.exit(1)

    refTEFileHandler = open(refTEFileName, "r")
        
    try:
        CheckerUtils.checkHeaders(refTEFileHandler)
    except CheckerException, e:
        
        print "Error in file " + refTEFileName + ". Wrong headers are :"
        print separator.join(e.messages)
        print "Authorized characters are : a-z A-Z 0-9 - . : _ | =\n"
        inGenomeFileHandler.close()
        sys.exit(1)
            
    refTEFileHandler.close()



    if os.path.exists( "%s_db" % ( projectName ) ):
        os.system( "rm -rf %s_db" % ( projectName ) )
    os.mkdir( "%s_db" % ( projectName ) )
    os.chdir( "%s_db" % ( projectName ) )

    #--------------------------------------------------------------------------

    if not os.path.exists( "%s/%s_db/%s.fa" % ( projectDir, projectName, projectName ) ):
        os.system( "ln -s ../%s.fa %s.fa" % ( projectName, projectName ) )
        
    isExceptionRaised = False
    sectionName = "prepare_data"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName)
    except NoSectionError:
        print "ERROR: the section " + sectionName + " must be in your config file : " + configFileName
        sys.exit(1)
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "chunk_length") 
    except NoOptionError:
        print "ERROR: the option chunk_length must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "chunk_overlap") 
    except NoOptionError:
        print "ERROR: the option chunk_overlap must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "nb_seq_per_batch") 
    except NoOptionError:
        print "ERROR: the option nb_seq_per_batch must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "make_random_chunks") 
    except NoOptionError:
        print "ERROR: the option make_random_chunks must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    if isExceptionRaised: 
        sys.exit(1)

    # load the chromosome sequences into a MySQL table
    prg = os.environ["REPET_PATH"] + "/bin/srptCreateTable.py"
    cmd = prg
    cmd += " -f %s.fa" % ( projectName )
    cmd += " -n %s_chr_seq" % ( projectName )
    cmd += " -t fasta"
    cmd += " -v %i" % ( verbose - 1 )
    pL.launch( prg, cmd, verbose - 1 )

    # cut the genomic sequences into chunks
    prg = os.environ["REPET_PATH"] + "/bin/dbChunks.py"
    cmd = prg
    cmd += " -i %s.fa" % ( projectName )
    cmd += " -l " + config.get("prepare_data","chunk_length")
    cmd += " -o " + config.get("prepare_data","chunk_overlap")
    cmd += " -w 0"
    cmd += " -O %s_chunks" % ( projectName )
    cmd += " -c"
    cmd += " -v %i" % ( verbose - 1 )
    pL.launch( prg, cmd, verbose - 1 )

    # load the chunk sequences into a MySQL table
    prg = os.environ["REPET_PATH"] + "/bin/srptCreateTable.py"
    cmd = prg
    cmd += " -f %s_chunks.fa" % ( projectName )
    cmd += " -n %s_chk_seq" % ( projectName )
    cmd += " -t fasta"
    cmd += " -v %i" % ( verbose - 1 )
    pL.launch( prg, cmd, verbose - 1 )

    # load the links between chunks and chromosomes into a MySQL table
    prg = os.environ["REPET_PATH"] + "/bin/srptCreateTable.py"
    cmd = prg
    cmd += " -f %s_chunks.map" % ( projectName )
    cmd += " -n %s_chk_map" % ( projectName )
    cmd += " -t map"
    cmd += " -v %i" % ( verbose - 1 )
    pL.launch( prg, cmd, verbose - 1 )

    # allocate one (or several) chunks(s) into the same file (-> batch)
    prg = os.environ["REPET_PATH"] + "/bin/dbSplit.py"
    cmd = prg
    cmd += " -i %s_chunks.fa" % ( projectName )
    cmd += " -n %s" % ( config.get("prepare_data","nb_seq_per_batch") )
    cmd += " -d"
    cmd += " -v %i" % ( verbose - 1 )
    pL.launch( prg, cmd, verbose - 1 )

    # shuffle each chunk independently to get randomized chunks
    if config.get( "prepare_data", "make_random_chunks" ) == "yes":
        if os.path.exists( "batches_rnd" ):
            os.rmdir( "batches_rnd" )
        prg = os.environ["REPET_PATH"] + "/bin/dbShuffle.py"
        cmd = prg
        cmd += " -I batches"
        cmd += " -O batches_rnd"
        cmd += " -v %i" % ( verbose - 1 )
        pL.launch( prg, cmd, verbose - 1 )

    #--------------------------------------------------------------------------

    if not os.path.exists( "%s/%s_db/%s_refTEs.fa" % ( projectDir, projectName, projectName ) ):
        os.system( "ln -s ../%s_refTEs.fa %s_refTEs.fa_initH" % ( projectName, projectName ) )

    # load the TE library into a MySQL table
    prg = os.environ["REPET_PATH"] + "/bin/srptCreateTable.py"
    cmd = prg
    cmd += " -f %s_refTEs.fa_initH" % ( projectName )
    cmd += " -n %s_refTEs_seq" % ( projectName )
    cmd += " -t fasta"
    cmd += " -v %i" % ( verbose - 1 )
    pL.launch( prg, cmd, verbose - 1 )

    # shorten sequence headers
    prg = os.environ["REPET_PATH"] + "/bin/shortenSeqHeader.py"
    cmd = prg
    cmd += " -i %s_refTEs.fa_initH" % ( projectName )
    cmd += " -l 1"
    cmd += " -p refTE_"
    cmd += " -o %s_refTEs.fa" % ( projectName )
    cmd += " -v %i" % ( verbose - 1 )
    pL.launch( prg, cmd, verbose - 1 )

    # load the links between initial headers and new headers into a MySQL table
    prg = os.environ["REPET_PATH"] + "/bin/srptCreateTable.py"
    cmd = prg
    cmd += " -f %s_refTEs.fa_initH.shortHlink" % ( projectName )
    cmd += " -n %s_refTEs_map" % ( projectName )
    cmd += " -t map"
    cmd += " -v %i" % ( verbose - 1 )
    pL.launch( prg, cmd, verbose - 1 )

    # prepare the TE library for BLASTER
    if config.get("align_refTEs_with_genome","blast") == "ncbi":
        if not CheckerUtils.isExecutableInUserPath("formatdb"):
            print "ERROR: NCBI-BLAST formatdb must be in your path"
            sys.exit(1)
    elif config.get("align_refTEs_with_genome","blast") == "wu":
        if not CheckerUtils.isExecutableInUserPath("xdformat"):
            print "ERROR: WU-BLAST xdformat must be in your path"
            sys.exit(1)
    prg = os.environ["REPET_PATH"] + "/bin/blaster"
    cmd = prg
    cmd += " -q %s_refTEs.fa" % ( projectName )
    cmd += " -B %s_refTEs.fa-BLRnPrepare" % ( projectName )
    cmd += " -P"
    if config.get("align_refTEs_with_genome","blast") == "wu":
        cmd += " -W"
    pL.launch( prg, cmd, verbose - 1 )

    os.chdir( ".." )

    if verbose > 0:
        print "step 1 finished successfully\n"; sys.stdout.flush()

#------------------------------------------------------------------------------

def alignChunksAndTEs( qryDirSuffix, align ):

    """
    Step 2: map the sequence from the TE library on the genomic chunks

    @param qryDirSuffix: query directory suffix ('' or '_rnd')
    @type qryDirSuffix: string

    @param align: alignment program (Blaster, RepeatMasker or Censor)
    @type align: string
    """

    if verbose > 0:
        print "\nbeginning of step 2"
        if qryDirSuffix == "":
            print "use %s to align the TE sequences on the chunks" % ( align )
        else:
            print "use %s to align the TE sequences on the randomized chunks" % ( align )
            if not os.path.exists( projectDir + "/%s_db/batches_rnd" % ( projectName ) ):
                print "ERROR: you didn't build random chunks at step 1, check the configuration file"
                sys.exit(1)
        sys.stdout.flush()

    if not os.path.exists( "%s_TEdetect%s" % ( projectName, qryDirSuffix ) ):
        os.mkdir( "%s_TEdetect%s" % ( projectName, qryDirSuffix ) )
    os.chdir( "%s_TEdetect%s" % ( projectName, qryDirSuffix ) )
    if not os.path.exists( align ):
        os.mkdir( align )
    else:
        print "ERROR: directory %s_TEdetect%s/%s/ already exists" % ( projectName, qryDirSuffix, align )
        sys.exit(1)
    os.chdir( align )

    # launch the alignment program (BLR+M or RM or CEN) on all the chunks in parallel
    if align == "RM":
        if not CheckerUtils.isExecutableInUserPath("RepeatMasker"):
            print "ERROR: RepeatMasker must be in your path"
            sys.exit(1)
    if align == "CEN":
        if not CheckerUtils.isExecutableInUserPath("censor"):
            print "ERROR: censor must be in your path"
            sys.exit(1)
        
    isExceptionRaised = False
    sectionName = "parallelized_jobs"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName)
    except NoSectionError:
        print "ERROR: the section " + sectionName + " must be in your config file : " + configFileName
        sys.exit(1)
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "queue") 
    except NoOptionError:
        print "ERROR: the option queue must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "tmpDir") 
    except NoOptionError:
        print "ERROR: the option tmpDir must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "clean") 
    except NoOptionError:
        print "ERROR: the option clean must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    if isExceptionRaised: 
        sys.exit(1)
        
    cmdEnd = " -g %s_TEannot_%s%s" % ( projectName, align, qryDirSuffix )
    cmdEnd += " -t jobs"
    cmdEnd += " -q %s/%s_db/batches%s" % ( projectDir, projectName, qryDirSuffix )
    cmdEnd += " -s %s/%s_db/%s_refTEs.fa" % ( projectDir, projectName, projectName )
    cmdEnd += " -Q \"%s\"" % ( config.get( "parallelized_jobs", "queue" ) )
    if config.get("parallelized_jobs","tmpDir") != "":
        cmdEnd += " -d %s" % ( config.get("parallelized_jobs","tmpDir") )
    if config.get( "parallelized_jobs", "clean" ) == "yes":
        cmdEnd += " -c"
    if os.environ["REPET_JOBS"] == "files":
        cmdEnd += " -p " + projectDir
    if align == "BLR":
        
        sectionName = "align_refTEs_with_genome"
        try:
            CheckerUtils.checkSectionInConfigFile(config, sectionName)
        except NoSectionError:
            print "ERROR: the section " + sectionName + " must be in your config file : " + configFileName
            sys.exit(1)
            
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "blast") 
        except NoOptionError:
            print "ERROR: the option 'blast' must be define in " + sectionName +" in your config file : " + configFileName
            sys.exit(1)
            
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "BLR_sensitivity") 
        except NoOptionError:
            print "ERROR: the option BLR_sensitivity must be define in " + sectionName +" in your config file : " + configFileName
            sys.exit(1)
            
        prg = os.environ["REPET_PATH"] + "/bin/srptBlasterMatcher.py"
        cmd = prg
        cmd += " -m 1"
        cmd += " -B \"-S %s" % ( config.get("align_refTEs_with_genome","BLR_sensitivity") )
        if config.get("align_refTEs_with_genome","blast") == "wu":
            cmd += " -W"
        cmd += "\""
        cmd += cmdEnd
    elif align == "RM":
        prg = os.environ["REPET_PATH"] + "/bin/srptRM.py"
        cmd = prg
        cmd += " -P \"-s -cutoff 200"
        if config.get("align_refTEs_with_genome","blast") == "wu":
            cmd += " -e wublast"
        cmd += "\""
        cmd += cmdEnd
    elif align == "CEN":
        prg = os.environ["REPET_PATH"] + "/bin/srptCEN.py"
        cmd = prg
        cmd += " -P \"-s -ns\""
        cmd += cmdEnd
    else:
        print "ERROR: option -a %s not recognized" % ( align )
        sys.exit(1)
        
    if os.environ[ "REPET_QUEUE" ] == "SGE":
        if not CheckerUtils.isQueueNameValid(config.get("parallelized_jobs","queue")):
            print "ERROR: the queue name specified in parallelized_jobs section in your configuration file is not valid"
            sys.exit(1)
            
    pL.launch( prg, cmd, verbose - 1 )

    os.chdir( "../.." )

    if verbose > 0:
        print "step 2 finished successfully\n"; sys.stdout.flush()

#------------------------------------------------------------------------------

def filterAndCombineTEsAnnot( combAlign ):

    """
    Step 3: filter the true HSPs with the randomized ones and combine them among the 3 programs
    """

    if verbose > 0:
        print "\nbeginning of step 3"
        print "filter the real HSPs with the random HSPs and combine them"
        sys.stdout.flush()

    if combAlign == "":
        print "ERROR: no alignment program specified with option -c"
        sys.exit(1)

    isExceptionRaised = False
    sectionName = "parallelized_jobs"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName)
    except NoSectionError:
        print "ERROR: the section " + sectionName + " must be in your config file : " + configFileName
        sys.exit(1)
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "queue") 
    except NoOptionError:
        print "ERROR: the option queue must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "tmpDir") 
    except NoOptionError:
        print "ERROR: the option tmpDir must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "clean") 
    except NoOptionError:
        print "ERROR: the option clean must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    if isExceptionRaised: 
        sys.exit(1)

    # retrieve for each program ran on the random chunks the highest score from all 'align' files
    thres = {}
    
    isExceptionRaised = False
    sectionName = "filter"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName)
    except NoSectionError:
        print "ERROR: the section " + sectionName + " must be in your config file : " + configFileName
        sys.exit(1)
                    
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "force_default_values") 
    except NoOptionError:
        print "ERROR: the option force_default_values must be define in " + sectionName + " in your config file : " + configFileName
        isExceptionRaised = True
       
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "BLR") 
    except NoOptionError:
        print "ERROR: the option BLR must be define in " + sectionName + " in your config file : " + configFileName
        isExceptionRaised = True
       
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "RM") 
    except NoOptionError:
        print "ERROR: the option RM must be define in " + sectionName + " in your config file : " + configFileName
        isExceptionRaised = True
       
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "CEN") 
    except NoOptionError:
        print "ERROR: the option CEN must be define in " + sectionName + " in your config file : " + configFileName
        isExceptionRaised = True

    if isExceptionRaised: 
        sys.exit(1)

    if os.environ[ "REPET_QUEUE" ] == "SGE":
        if not CheckerUtils.isQueueNameValid(config.get("parallelized_jobs","queue")):
            print "ERROR: the queue name specified in parallelized_jobs section in your configuration file is not valid"
            sys.exit(1)

    if os.path.exists( projectName + "_TEdetect_rnd" ):
        os.chdir( projectName + "_TEdetect_rnd" )
        if verbose > 0:
            print "\n* retrieve for each program ran on the random chunks the highest score from all 'align' files"; sys.stdout.flush()
        scoresFile = "maxScores.txt"
        scoresFileHandler = open( scoresFile, "w" )
        scoresFileHandler.write( "algo\tfile\tmaxScore\n" )
        for a in ["BLR","RM","CEN"]:
            if not os.path.exists( a ):
                if verbose > 0:
                    print "WARNING: %s doesn't exist, did you run the previous step ?" % ( a ); sys.stdout.flush()
            else:
                lFiles = glob.glob( "%s/*.align" % ( a ) )
                lFiles.sort()
                lHighestScores = []
                for fileName in lFiles:
                    lAlignInstances = AlignUtils.getAlignListFromFile( fileName )
                    lScores = AlignUtils.getListOfScores( lAlignInstances )
                    if lScores != []:
                        lHighestScores.append( max( lScores ) )
                        scoresFileHandler.write( "%s\t%s\t%i\n" % ( a, os.path.basename(fileName), max(lScores) ) )
                    else:
                        lHighestScores.append( 0 )
                        scoresFileHandler.write( "%s\t%s\t0\n" % ( a, os.path.basename(fileName) ) )
                iStat = Stat( lHighestScores )
                if verbose > 0:
                    iStat.viewQuantiles()
                thres[a] = str( iStat.quantile( 0.95 ) )
                if thres[a] != "":
                    if verbose > 0:
                        print "threshold %s: %s" % ( a, thres[a] ); sys.stdout.flush()
                else:
                    thres[a] = config.get( "filter", a )
                    if verbose > 0:
                        print "WARNING: no threshold found for %s. Default value is %s." % ( a, thres[a] ); sys.stdout.flush()
        scoresFileHandler.close()

    if config.get( "filter", "force_default_values" ) == "yes":
        if verbose > 0:
            print "\n* force the use of the default values from the configuration file"; sys.stdout.flush()
            for a in ["BLR","RM","CEN"]:
                thres[a] = config.get( "filter", a )
                if verbose > 0:
                    print "default threshold " + a + ": " + thres[a]; sys.stdout.flush()

    if thres == {}:
        print "ERROR: you have to launch step 2 with option \"-r\" or specify default filter values in the configuration file"
        sys.exit(1)

    os.chdir( projectDir + "/" + projectName + "_TEdetect/" )

    # filter the 'align' files
    if verbose > 0:
        print "\n* filter the 'align' files according to the threshold"; sys.stdout.flush()
    for prg in ["BLR","RM","CEN"]:
        if os.path.exists( prg ):
                os.chdir( prg )
                lAlignFiles = glob.glob( "*.align" )
                for alignFile in lAlignFiles:
                    cmd = "gawk -v thres=" + thres[prg] + " '$8>thres{print $0}' " + alignFile + " > " + alignFile + ".filtered"
                    log = os.system( cmd )
                    if log != 0:
                        print "ERROR while filtering '%s' from %s" % ( alignFile, prg )
                        sys.exit(1)
                os.chdir( ".." )

    # concatenate chunk by chunk the filtered 'align' files from each alignment program
    if verbose > 0:
        print "\n* concatenate chunk by chunk the filtered 'align' files from each alignment program"; sys.stdout.flush()
    if os.path.exists( "Comb" ):
        os.system( "rm -rf Comb" )
    os.mkdir( "Comb" )
    os.chdir( "Comb" )
    lChkFilePaths = glob.glob( "%s/%s_db/batches/batch*.fa" % ( projectDir, projectName ) )
    for chkFilePath in lChkFilePaths:
        chkFileName = os.path.basename( chkFilePath )
        prg = "cat"
        cmd = prg
        if "BLR" in combAlign and os.path.exists( "../BLR/%s.align.filtered" % chkFileName ):
            if "RM" in combAlign and os.path.exists( "../RM/%s.cat.align.filtered" % chkFileName ):
                if "CEN" in combAlign and os.path.exists( "../CEN/%s.map.align.filtered" % chkFileName ):
                    cmd += " ../BLR/%s.align.filtered ../RM/%s.cat.align.filtered ../CEN/%s.map.align.filtered > %s_allTEs.align" % ( chkFileName, chkFileName, chkFileName, chkFileName )
                elif "CEN" not in combAlign or not os.path.exists( "../CEN/%s.map.align.filtered" % chkFileName ):
                    cmd += " ../BLR/%s.align.filtered ../RM/%s.cat.align.filtered > %s_allTEs.align" % ( chkFileName, chkFileName, chkFileName )
            elif "RM" not in combAlign or not os.path.exists( "../RM/%s.cat.align.filtered" % chkFileName ):
                if "CEN" in combAlign and os.path.exists( "../CEN/%s.map.align.filtered" % chkFileName ):
                    cmd += " ../BLR/%s.align.filtered ../CEN/%s.map.align.filtered > %s_allTEs.align" % ( chkFileName, chkFileName, chkFileName )
                elif "CEN" not in combAlign or not os.path.exists( "../CEN/%s.map.align.filtered" % chkFileName ):
                    cmd += " ../BLR/%s.align.filtered > %s_allTEs.align" % ( chkFileName, chkFileName )
        elif "BLR" not in combAlign or not os.path.exists( "../BLR/%s.align.filtered" % chkFileName ):
            if "RM" in combAlign and os.path.exists( "../RM/%s.cat.align.filtered" % chkFileName ):
                if "CEN" in combAlign and os.path.exists( "../CEN/%s.map.align.filtered" % chkFileName ):
                    cmd += " ../RM/%s.cat.align.filtered ../CEN/%s.map.align.filtered > %s_allTEs.align" % ( chkFileName, chkFileName, chkFileName )
                elif "CEN" not in combAlign or not os.path.exists( "../CEN/%s.map.align.filtered" % chkFileName ):
                    cmd += " ../RM/%s.cat.align.filtered > %s_allTEs.align" % ( chkFileName, chkFileName )
            elif "RM" not in combAlign or not os.path.exists( "../RM/%s.cat.align.filtered" % chkFileName ):
                if "CEN" in combAlign and os.path.exists( "../CEN/%s.map.align.filtered" % chkFileName ):
                    cmd += " ../CEN/%s.map.align.filtered > %s_allTEs.align" % ( chkFileName, chkFileName )
                elif "CEN" not in combAlign or not os.path.exists( "../CEN/%s.map.align.filtered" % chkFileName ):
                    continue
        pL.launch( prg, cmd, verbose - 1 )

    # join the matches with Matcher
    string = "join the matches with Matcher"
    if verbose > 0:
        print "\n* %s" % ( string ); sys.stdout.flush()
    prg = os.environ["REPET_PATH"] + "/bin/srptBlasterMatcher.py"
    cmd = prg
    cmd += " -g %s_TEannot_Matcher" % ( projectName )
    cmd += " -q %s/%s_TEdetect/Comb" % ( projectDir, projectName )
    cmd += " -Q \"%s\"" % ( config.get( "parallelized_jobs", "queue" ) )
    cmd += " -m 2"   # launch Matcher only
    cmd += " -M \"-j -x\""
    cmd += " -Z path"
    cmd += " -C %s/%s" % ( projectDir, configFileName )
    if config.get("parallelized_jobs","clean") == "yes":
        cmd += " -c"
    cmd += " -v %i" % ( verbose - 1 )
    if config.get("parallelized_jobs","tmpDir") != "":
        cmd += " -d %s" % ( config.get("parallelized_jobs","tmpDir") )
    pL.launch( prg, cmd, verbose - 1 )
    
    # load data in MySQL
    prg = os.environ["REPET_PATH"] + "/bin/srptCreateTable.py"
    cmd = prg
    cmd += " -f %s_TEannot_Matcher.path" % ( projectName )
    cmd += " -n %s_chk_allTEs_tmpH_path" % ( projectName )
    cmd += " -t path"
    cmd += " -v %i" % ( verbose - 1 )
    pL.launch( prg, cmd, verbose - 1 )

    # retrieve the initial headers of the reference TEs
    prg = os.environ["REPET_PATH"] + "/bin/srptRetrieveInitHeaders.py"
    cmd = prg
    cmd += " -n %s_chk_allTEs_tmpH_path" % ( projectName )
    cmd += " -l %s/%s_db/%s_refTEs.fa_initH.shortHlink" % ( projectDir, projectName, projectName )
    cmd += " -s"
    cmd += " -C %s/%s" % ( projectDir, configFileName )
    cmd += " -o %s_chk_allTEs_path" % ( projectName )
    cmd += " -c"
    pL.launch( prg, cmd, verbose - 1 )

    # convert the coordinates from chunks to chromosomes
    prg = os.environ["REPET_PATH"] + "/bin/srptConvChunk2Chr.py"
    cmd = prg
    cmd += " -m %s_chk_map" % ( projectName )
    cmd += " -q %s_chk_allTEs_path" % ( projectName )
    cmd += " -t path"
    cmd += " -c"
    cmd += " -o %s_chr_allTEs_path" % ( projectName )
    cmd += " -v %i" % ( verbose - 1 )
    pL.launch( prg, cmd, verbose - 1 )

    # clean
    if config.get("parallelized_jobs","clean") == "yes":
        cmd = "find . -name \"*.align\" -exec rm {} \;"
        log = os.system( cmd )
        if log != 0:
            print "ERROR while removing temporary 'align' files"
            sys.exit(1)
        cmd = "find . -name \"*.align.clean_match.map\" -exec rm {} \;"
        log = os.system( cmd )
        if log != 0:
            print "ERROR while removing temporary 'map' files"
            sys.exit(1)

    os.chdir( "../.." )

    if verbose > 0:
        print "step 3 finished successfully\n"; sys.stdout.flush()

#------------------------------------------------------------------------------

def detectSSRs( sat ):

    """
    Step 4: detect SSRs

    @param sat: SSR detection program (TRF, RMSSR, Mreps)
    @type sat: string
    """

    if verbose > 0:
        print "\nbeginning of step 4"
        print "search for satellites on the chunks with %s" %( sat )
        sys.stdout.flush()

    if not os.path.exists( "%s_SSRdetect" % ( projectName ) ):
        os.mkdir( "%s_SSRdetect" % ( projectName ) )
    os.chdir( "%s_SSRdetect" % ( projectName ) )
    if not os.path.exists( sat ):
        os.mkdir( sat )
    else:
        print "ERROR: directory %s_SSRdetect/%s/ already exists" % ( projectName, sat )
        sys.exit(1)
    os.chdir( sat )

    # launch the SSR-detection program (TRF or Mreps or RM) on  all the chunks in parallel
    if sat ==  "TRF" and not CheckerUtils.isExecutableInUserPath("trf"):
        print "ERROR: trf must be in your path"
        sys.exit(1)
    if sat ==  "RMSSR" and not CheckerUtils.isExecutableInUserPath("RepeatMasker"):
        print "ERROR: RepeatMasker must be in your path"
        sys.exit(1)
    if sat ==  "Mreps" and not CheckerUtils.isExecutableInUserPath("mreps"):
        print "ERROR: mreps must be in your path"
        sys.exit(1)

    isExceptionRaised = False
    sectionName = "parallelized_jobs"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName)
    except NoSectionError:
        print "ERROR: the section " + sectionName + " must be in your config file : " + configFileName
        sys.exit(1)
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "queue") 
    except NoOptionError:
        print "ERROR: the option queue must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "tmpDir") 
    except NoOptionError:
        print "ERROR: the option tmpDir must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "clean") 
    except NoOptionError:
        print "ERROR: the option clean must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    if isExceptionRaised: 
        sys.exit(1)
        
    if os.environ[ "REPET_QUEUE" ] == "SGE":
        if not CheckerUtils.isQueueNameValid(config.get("parallelized_jobs","queue")):
            print "ERROR: the queue name specified in parallelized_jobs section in your configuration file is not valid"
            sys.exit(1)    

    cmdEnd = " -g %s_TEannot_%s" % ( projectName, sat )
    cmdEnd += " -t jobs"
    cmdEnd += " -q %s/%s_db/batches" % ( projectDir, projectName )
    cmdEnd += " -Q \"%s\"" % ( config.get( "parallelized_jobs", "queue" ) )
    cmdEnd += " -C"
    if config.get( "parallelized_jobs", "clean" ) == "yes":
        cmdEnd += " -c"
    if config.get("parallelized_jobs","tmpDir") != "":
        cmdEnd += " -d %s" % ( config.get("parallelized_jobs","tmpDir") )
    if sat == "TRF":
        prg = os.environ["REPET_PATH"] + "/bin/srptTRF.py"
        cmd = prg
        cmd += cmdEnd
    elif sat == "RMSSR":
        prg = os.environ["REPET_PATH"] + "/bin/srptRMSSR.py"
        cmd = prg
        cmd += " -P \"-s"
        if config.get("align_refTEs_with_genome","blast") == "wu":
            cmd += " -e wublast"
        cmd += "\""
        cmd += cmdEnd
    elif sat == "Mreps":
        prg = os.environ["REPET_PATH"] + "/bin/srptMreps.py"
        cmd = prg
        cmd += cmdEnd
    else:
        print "ERROR: option -s %s not recognized" % ( sat )
        sys.exit(1)
    pL.launch( prg, cmd, verbose - 1 )

    # load the results into a MySQL table
    prg = os.environ["REPET_PATH"] + "/bin/srptCreateTable.py"
    cmd = prg
    if sat == "TRF":
        cmd += " -f %s_TEannot_%s.set" % ( projectName, sat )
        cmd += " -n %s_chk_%s_set" % ( projectName, sat )
        cmd += " -t set"
    elif sat == "RMSSR":
        cmd += " -f %s_TEannot_%s.path" % ( projectName, sat )
        cmd += " -n %s_chk_%s_path" % ( projectName, sat )
        cmd += " -t path"
    elif sat == "Mreps":
        cmd += " -f %s_TEannot_%s.set" % ( projectName, sat )
        cmd += " -n %s_chk_%s_set" % ( projectName, sat )
        cmd += " -t set"
    cmd += " -v %i" % ( verbose - 1 )
    pL.launch( prg, cmd, verbose - 1 )

    os.chdir( ".." )

    if verbose > 0:
        print "step 4 finished successfully\n"; sys.stdout.flush()

#------------------------------------------------------------------------------

def filterAndCombineSSRsAnnot():

    """
    Step 5: merge the SSR annotations from TRF, Mreps and RepeatMasker
    """

    if verbose > 0:
        print "\nbeginning of step 5"
        print "filter and combine the SSR annotations from TRF, Mreps and RepeatMasker"
        sys.stdout.flush()

    if not os.path.exists( "%s_SSRdetect" % ( projectName ) ):
        print "ERROR: directory %s_SSRdetect doesn't exist" % ( projectName )
        sys.exit(1)
    os.chdir( "%s_SSRdetect" % ( projectName ) )
    if os.path.exists( "Comb" ):
        print "ERROR: directory %s_SSRdetect/Comb already exists" % ( projectName )
        sys.exit(1)
    os.mkdir( "Comb" )
    os.chdir( "Comb" )

    mergeFile = open( "mergeSSR.txt", "w" )
    cmd = "%s_chk_TRF_set\tset\tmerge\n" % ( projectName )
    cmd += "%s_chk_Mreps_set\tset\tmerge\n" % ( projectName )
    cmd += "%s_chk_RMSSR_path\tpath\tmerge\n" % ( projectName )
    mergeFile.write( cmd )
    mergeFile.close()

    prg = os.environ["REPET_PATH"] + "/bin/srptAutoPromote.py"
    cmd = prg
    cmd += " -f mergeSSR.txt"
    cmd += " -o %s_chk_allSSRs_set" % ( projectName )
    cmd += " -v %i" % ( verbose - 1 )
    pL.launch( prg, cmd, verbose - 1 )

    # convert the coordinates from chunks to chromosomes
    prg = os.environ["REPET_PATH"] + "/bin/srptConvChunk2Chr.py"
    cmd = prg
    cmd += " -m %s_chk_map" % ( projectName )
    cmd += " -q %s_chk_allSSRs_set" % ( projectName )
    cmd += " -t set"
    cmd += " -c"
    cmd += " -o %s_chr_allSSRs_set" % ( projectName )
    cmd += " -v %i" % ( verbose - 1 )
    pL.launch( prg, cmd, verbose - 1 )

    os.chdir( "../.." )

    if verbose > 0:
        print "step 5 finished successfully\n"; sys.stdout.flush()

#------------------------------------------------------------------------------

def alignChunksAndTEsForCoding( typeBlast ):

    """
    Step 6: align the 2nd data bank on the genomic sequences via Blaster with tblastx or blastx
    * load the bank in a MySQL table
    * prepare the bank
    * launch BLASTER with tblastx or blastx
    """

    if verbose > 0:
        print "\nbeginning of step 6"; sys.stdout.flush()

    #--------------------------------------------------------------------------

    if not os.path.exists( "%s_db" % ( projectName ) ):
        os.mkdir( "%s_db" % ( projectName ) )
    os.chdir( "%s_db" % ( projectName ) )
    
    if typeBlast == "tblastx":
        bankBlast = "bankBLRtx"
    elif typeBlast == "blastx":
        bankBlast = "bankBLRx"

    sectionName = "align_other_banks"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName)
    except NoSectionError:
        print "ERROR: the section " + sectionName + " must be in your config file : " + configFileName
        sys.exit(1)
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, bankBlast) 
    except NoOptionError:
        print "ERROR: the option " + bankBlast + " must be define in " + sectionName +" in your config file : " + configFileName
        sys.exit(1)
    isExceptionRaised = False
    sectionName = "parallelized_jobs"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName)
    except NoSectionError:
        print "ERROR: the section " + sectionName + " must be in your config file : " + configFileName
        sys.exit(1)
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "queue") 
    except NoOptionError:
        print "ERROR: the option queue must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "tmpDir") 
    except NoOptionError:
        print "ERROR: the option tmpDir must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "clean") 
    except NoOptionError:
        print "ERROR: the option clean must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    if isExceptionRaised: 
        sys.exit(1)
        
    if os.environ[ "REPET_QUEUE" ] == "SGE":
        if not CheckerUtils.isQueueNameValid(config.get("parallelized_jobs","queue")):
            print "ERROR: the queue name specified in parallelized_jobs section in your configuration file is not valid"
            sys.exit(1)

    bankName = config.get( "align_other_banks", bankBlast )
    if not os.path.exists( "%s/%s" % ( projectDir, bankName ) ):
        print "ERROR: %s/%s doesn't exist" % ( projectDir, bankName )
        sys.exit(1)

    if not os.path.exists( "%s/%s_db/%s" % ( projectDir, projectName, bankName ) ):
        os.system( "ln -s ../%s %s" % ( bankName, bankName ) )

    # load the bank in a MySQL table
    prg = os.environ["REPET_PATH"] + "/bin/srptCreateTable.py"
    cmd = prg
    cmd += " -f %s" % ( bankName )
    cmd += " -n %s_%s_nt_seq" % ( projectName, bankBlast )
    cmd += " -t fasta"
    cmd += " -v %i" % ( verbose - 1 )
    pL.launch( prg, cmd, verbose - 1 )

    # prepare the bank
    pL.reset( bankName )
    pL.launchBlaster( prepareOnly="yes", blastAlgo=typeBlast, outPrefix="%s-%sPrepare" % ( bankName, bankBlast ) )

    os.chdir( ".." )

    #--------------------------------------------------------------------------

    if verbose > 0:
        print "\nalign '%s' and the chunks via Blaster with %s" % ( bankName, typeBlast )
        sys.stdout.flush()

    if not os.path.exists( projectName + "_TEdetect" ):
        os.mkdir( projectName + "_TEdetect" )
    os.chdir( projectName + "_TEdetect" )
    if not os.path.exists( bankBlast ):
        os.mkdir( bankBlast )
    else:
        print "ERROR: directory %s_TEdetect/%s/ already exists" % ( projectName, bankBlast )
        sys.exit(1)
    os.chdir( bankBlast )

    # launch Blaster with tblastx followed by Matcher on each chunk in parallel
    prg = os.environ["REPET_PATH"] + "/bin/srptBlasterMatcher.py"
    cmd = prg
    cmd += " -g %s_TEannot_%s" % ( projectName, bankBlast )
    cmd += " -t jobs"
    cmd += " -q %s/%s_db/batches" % ( projectDir, projectName )
    cmd += " -s %s/%s_db/%s" % ( projectDir, projectName, bankName )
    cmd += " -m 3"
    cmd += " -B \"-n %s -r\"" % ( typeBlast )
    cmd += " -M \"-j\""
    cmd += " -Q \"%s\"" % ( config.get( "parallelized_jobs", "queue" ) )
    cmd += " -Z path"
    if config.get( "parallelized_jobs", "clean" ) == "yes":
        cmd += " -c"
    if config.get("parallelized_jobs","tmpDir") != "":
        cmd += " -d %s" % ( config.get("parallelized_jobs","tmpDir") )
    if os.environ["REPET_JOBS"] == "files":
        cmd += " -p " + projectDir
    pL.launch( prg, cmd, verbose - 1 )

    os.system( "mv %s_TEannot_%s.path %s_%s.path" % ( projectName, bankBlast, projectName, bankBlast ) )

    # load data in MySQL
    prg = os.environ["REPET_PATH"] + "/bin/srptCreateTable.py"
    cmd = prg
    cmd += " -f %s_%s.path" % ( projectName, bankBlast )
    cmd += " -n %s_chk_%s_path" % ( projectName, bankBlast )
    cmd += " -t path"
    cmd += " -v %i" % ( verbose - 1 )
    pL.launch( prg, cmd, verbose - 1 )

    # convert the coordinates from chunks to chromosomes
    prg = os.environ["REPET_PATH"] + "/bin/srptConvChunk2Chr.py"
    cmd = prg
    cmd += " -m %s_chk_map" % ( projectName )
    cmd += " -q %s_chk_%s_path" % ( projectName, bankBlast )
    cmd += " -t path"
    cmd += " -c"
    cmd += " -o %s_chr_%s_path" % ( projectName, bankBlast )
    cmd += " -v %i" % ( verbose - 1 )
    pL.launch( prg, cmd, verbose - 1 )

    os.chdir( ".." )

    if verbose > 0:
        print "step 6 finished successfully\n"; sys.stdout.flush()

#------------------------------------------------------------------------------

def pathsProcessing( config ):

    """
    Step 7: remove path doublons, spurious paths and apply long join procedure

    @param config: configuration file handling
    @type config: file handling
    """

    if verbose > 0:
        print "\nbeginning of step 7"
        print "long join procedure"
        sys.stdout.flush()

    os.chdir( "%s_TEdetect" % ( projectName ) )

    isExceptionRaised = False
    sectionName = "annot_processing"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName)
    except NoSectionError:
        print "ERROR: the section " + sectionName + " must be in your config file : " + configFileName
        sys.exit(1)
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "min_size") 
    except NoOptionError:
        print "ERROR: the option min_size must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "join_id_tolerance") 
    except NoOptionError:
        print "ERROR: the option join_id_tolerance must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "join_max_gap_size") 
    except NoOptionError:
        print "ERROR: the option join_max_gap_size must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "join_max_mismatch_size") 
    except NoOptionError:
        print "ERROR: the option join_max_mismatch_size must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "join_TEinsert_cov") 
    except NoOptionError:
        print "ERROR: the option join_TEinsert_cov must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "join_overlap") 
    except NoOptionError:
        print "ERROR: the option join_overlap must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "join_minlength_split") 
    except NoOptionError:
        print "ERROR: the option join_minlength_split must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    if isExceptionRaised == True:
        sys.exit(1)

    # remove path doublons among the TE annotations
    prg = os.environ["REPET_PATH"] + "/bin/srptRemovePathDoublons.py"
    cmd = prg
    cmd += " -i %s_chk_allTEs_path" % ( projectName )
    cmd += " -o %s_chk_allTEs_nr_path" % ( projectName )
    cmd += " -v %i" % ( verbose - 1 )
    pL.launch( prg, cmd, verbose - 1 )

    # remove SSRs inside TEs
    prg = os.environ["REPET_PATH"] + "/bin/srptRemoveSpurious.py"
    cmd = prg
    cmd += " -q %s_chk_allTEs_nr_path" % ( projectName )
    cmd += " -s %s_chk_allSSRs_set" % ( projectName )
    cmd += " -t path/set"
    cmd += " -r %s" % ( config.get("annot_processing","min_size") )
    cmd += " -o %s_chk_allTEs_nr_noSSR_path" % ( projectName )
    cmd += " -v %i" % ( verbose - 1 )
    pL.launch( prg, cmd, verbose - 1 )

    # remove temporary table
    prg = os.environ["REPET_PATH"] + "/bin/srptRemoveTable.py"
    cmd = prg
    cmd += " -t %s_chk_allSSRs_set_bin" % ( projectName )
    cmd += " -v %i" % ( verbose - 1 )
    pL.launch( prg, cmd, verbose - 1 )

    # convert the coordinates from chunks to chromosomes
    prg = os.environ["REPET_PATH"] + "/bin/srptConvChunk2Chr.py"
    cmd = prg
    cmd += " -m %s_chk_map" % ( projectName )
    cmd += " -q %s_chk_allTEs_nr_noSSR_path" % ( projectName )
    cmd += " -t path"
    cmd += " -c"
    cmd += " -o %s_chr_allTEs_nr_noSSR_path" % ( projectName )
    cmd += " -v %i" % ( verbose - 1 )
    pL.launch( prg, cmd, verbose - 1 )

    # long join procedure to connect fragmented, nested TE copies
    ljft = LongJoinsForTEs()
    ljft.setInputTable( "%s_chr_allTEs_nr_noSSR_path" % ( projectName ) )
    ljft.setIdentityTolerance( config.get("annot_processing","join_id_tolerance") )
    ljft.setMaxGapLength( config.get("annot_processing","join_max_gap_size") )
    ljft.setMaxMismatchLength( config.get("annot_processing","join_max_mismatch_size") )
    ljft.setMinNestedTEcoverage( config.get("annot_processing","join_TEinsert_cov") )
    ljft.setMaxOverlap( config.get("annot_processing","join_overlap") )
    ljft.setMinLengthToSplit( config.get("annot_processing","join_minlength_split") )
    ljft.setMinLengthToKeepChain( config.get("annot_processing","min_size") )
    ljft.setOutputTable( "%s_chr_allTEs_nr_noSSR_join_path" % ( projectName ) )
    ljft.setConfigFileName( "%s/%s" % ( projectDir, configFileName ) )
    ljft.setVerbosityLevel( verbose )
    ljft.run()

    # remove temporary table:
    prg = os.environ["REPET_PATH"] + "/bin/srptRemoveTable.py"
    cmd = prg
    cmd += " -t %s_chr_allTEs_nr_noSSR_path" % ( projectName )
    cmd += " -v %i" % ( verbose - 1 )
    pL.launch( prg, cmd, verbose - 1 )

    # convert the coordinates from chromosomes to chunks
    prg = os.environ["REPET_PATH"] + "/bin/srptConvChr2Chunk.py"
    cmd = prg
    cmd += " -m %s_chk_map" % ( projectName )
    cmd += " -q %s_chr_allTEs_nr_noSSR_join_path" % ( projectName )
    cmd += " -t path"
    cmd += " -o %s_chk_allTEs_nr_noSSR_join_path" % ( projectName )
    pL.launch( prg, cmd, verbose - 1 )

    os.chdir( ".." )

    if verbose > 0:
        print "step 7 finished successfully\n"; sys.stdout.flush()

#------------------------------------------------------------------------------

def exportAnnotations( outFormat ):

    """
    Step 8: export files recording the annotations for each chunk

    @outFormat: gameXML or GFF3
    @type outFormat: string
    """

    if verbose > 0:
        print "\nbeginning of step 8"
        print "build %s files" % ( outFormat )
        sys.stdout.flush()

    isExceptionRaised = False
    sectionName = "export"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName)
    except NoSectionError:
        print "ERROR: the section " + sectionName + " must be in your config file : " + configFileName
        sys.exit(1)
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "sequences") 
    except NoOptionError:
        print "ERROR: the option sequences must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "add_SSRs") 
    except NoOptionError:
        print "ERROR: the option add_SSRs must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "add_tBx") 
    except NoOptionError:
        print "ERROR: the option add_tBx must be define in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    if isExceptionRaised == True:
        sys.exit(1)

    refSeq = config.get("export","sequences")
    if refSeq == "chromosomes":
        refSeq = "chr"
    elif refSeq == "chunks":
        refSeq = "chk"
    else:
        print "ERROR: '%s' not recognized"
        sys.exit(1)

    if not os.path.exists( "%s_%s%s" % ( projectName, outFormat, refSeq ) ):
        os.mkdir( "%s_%s%s" % ( projectName, outFormat, refSeq ) )
    os.chdir( "%s_%s%s" % ( projectName, outFormat, refSeq ) )

    # write the file with the annotation tables
    tableFileName = "annotation_tables.txt"
    tableFile = open( tableFileName, "w" )
    string = "%s_REPET_TEs\t" % ( projectName )
    string += "path\t"
    string += "%s_%s_allTEs_nr_noSSR_join_path\n" % ( projectName, refSeq )
    if config.get("export","add_SSRs") == "yes":
        if os.path.exists( "%s/%s_SSRdetect/Comb" % ( projectDir, projectName ) ): 
            string += "%s_REPET_SSRs\t" % ( projectName )
            string += "set\t"
            string += "%s_%s_allSSRs_set\n" % ( projectName, refSeq )
        else:
            print "WARNING: can't export SSR annotations, launch step 4 and 5 first"
    if config.get("export","add_tBx") == "yes":
        if os.path.exists( "%s/%s_TEdetect/bankBLRtx" % ( projectDir, projectName ) ):
            string += "%s_REPET_tblastx\t" % ( projectName )
            string += "path\t"
            string += "%s_%s_bankBLRtx_path\n" % ( projectName, refSeq )
        else:
            print "WARNING: can't export tblastx annotations, launch step 6 first"
    if config.get("export","add_Bx") == "yes":
        if os.path.exists( "%s/%s_TEdetect/bankBLRx" % ( projectDir, projectName ) ):
            string += "%s_REPET_blastx\t" % ( projectName )
            string += "path\t"
            string += "%s_%s_bankBLRx_path\n" % ( projectName, refSeq )
        else:
            print "WARNING: can't export blastx annotations, launch step 6 first"
    tableFile.write( string )
    tableFile.close()

    #--------------------------------------------------------------------------

    if outFormat == "GFF3":
        
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "gff3_chado") 
        except NoOptionError:
            print "ERROR: the option gff3_chado must be define in " + sectionName +" in your config file : " + configFileName
            sys.exit(1)

        # create the source GFF3 file(s), one per genomic sequence
        if verbose > 0:
            print "create source files..."; sys.stdout.flush()
        if not os.path.exists( "src_GFF3" ):
            os.mkdir( "src_GFF3" )
        os.chdir( "src_GFF3" )
        prg = os.environ["REPET_PATH"] + "/bin/srptGFF3Maker.py"
        cmd = prg
        if refSeq == "chr":
            cmd += " -f %s_chr_seq" % ( projectName )
        elif refSeq == "chk":
            cmd += " -f %s_chk_seq" % ( projectName )
        cmd += " -v %i" % ( verbose - 1 )
        pL.launch( prg, cmd, verbose - 1 )
        os.chdir( ".." )

        # add the annotations to the GFF3 files
        if verbose > 0:
            print "add annotations to source files..."; sys.stdout.flush()
        if os.path.exists( "annotations" ):
            os.system( "rm -rf annotations" )
        os.mkdir( "annotations" )
        os.chdir( "annotations" )
        lGff3FilePaths = glob.glob( "../src_GFF3/*.gff3" )
        for gff3FilePath in lGff3FilePaths:
            gff3FileName = os.path.basename( gff3FilePath )
            shutil.copy( gff3FilePath, gff3FileName )
            prg = os.environ["REPET_PATH"] + "/bin/srptGFF3Maker.py"
            cmd = prg
            cmd += " -t ../%s" % ( tableFileName )
            cmd += " -g %s" % ( gff3FileName )
            if config.get("export","gff3_chado") == "yes":
                cmd += " -c"
            cmd += " -v %i" % ( verbose - 1 )
            pL.launch( prg, cmd )

        os.chdir( ".." )
        os.system( "rm -rf src_GFF3/" )
        os.chdir( ".." )

    #--------------------------------------------------------------------------

    elif outFormat == "gameXML":

        # create the source gameXML file(s), one per genomic sequence
        if verbose > 0:
            print "create source files..."; sys.stdout.flush()
        if not os.path.exists( "src_gameXML" ):
            os.mkdir( "src_gameXML" )
        os.chdir( "src_gameXML" )

        if refSeq == "chr":
            if not os.path.exists( "%s.fa" % ( projectName ) ):
                os.system( "ln -s " + projectDir + "/" + projectName + ".fa " + projectName + ".fa" )
        elif refSeq == "chk":
            if not os.path.exists( "%s_chunks.fa" % ( projectName ) ):
                os.system( "ln -s " + projectDir + "/" + projectName + "_db/" + projectName + "_chunks.fa " + projectName + "_chunks.fa" )

        prg = os.environ["REPET_PATH"] + "/bin/srptGameXmlMaker.py"
        cmd = prg
        if refSeq == "chr":
            cmd += " -f %s.fa" % ( projectName )
        elif refSeq == "chk":
            cmd += " -f %s_chunks.fa" % ( projectName )
        cmd += " -v %i" % ( verbose - 1 )
        pL.launch( prg, cmd, verbose - 1 )

        if refSeq == "chr":
            os.remove( projectName + ".fa" )
        elif refSeq == "chk":
            os.remove( projectName + "_chunks.fa" )

        exitStatus = os.system( os.environ["REPET_PATH"] + "/bin/renameFiles.py -i gamexml.new -o gamexml.ori -l \"*.new\" > /dev/null 2>&1" )
        if exitStatus != 0:
            print "ERROR while renaming gameXML files ('.new'->'.ori')"; sys.exit(1)
        os.chdir( ".." )

        # add the annotations to the gameXML files
        if verbose > 0:
            print "add annotations to source files..."; sys.stdout.flush()
        if os.path.exists( "annotations" ):
            os.system( "rm -rf annotations" )
        os.mkdir( "annotations" )
        os.chdir( "annotations" )
        os.system( "cp ../src_gameXML/*.ori ." )
        exitStatus = os.system( os.environ["REPET_PATH"] + "/bin/renameFiles.py -i gamexml.ori -o gamexml -l \"*.ori\" > /dev/null 2>&1" )
        if exitStatus != 0:
            print "ERROR while renaming gameXML files ('.ori'->'.gamexml')"; sys.exit(1)
        lGXmlFilePaths = glob.glob( "*.gamexml" )
        for gXmlFilePath in lGXmlFilePaths:
            gXmlFileName = os.path.basename( gXmlFilePath )
            #shutil.copy( gXmlFilePath, gXmlFileName )
            prg = os.environ["REPET_PATH"] + "/bin/srptGameXmlMaker.py"
            cmd = prg
            cmd += " -t ../%s" % ( tableFileName )
            cmd += " -g %s" % ( gXmlFileName )
            cmd += " -v %i" % ( verbose - 1 )
            pL.launch( prg, cmd )
        exitStatus = os.system( os.environ["REPET_PATH"] + "/bin/renameFiles.py -i gamexml.new -o gamexml -l \"*.new\" > /dev/null 2>&1" )
        if exitStatus != 0:
            print "ERROR while renaming gameXML files ('.new'->'.gamexml')"; sys.stdout.flush()

        os.chdir( ".." )
        os.system( "rm -rf src_gameXML/" )
        os.chdir( ".." )

    if verbose > 0:
        print "step 8 finished successfully\n"; sys.stdout.flush()

#------------------------------------------------------------------------------

if __name__ == "__main__":
    main()
