#!/usr/bin/env python

"""
Pipeline for the de novo identification of transposable elements (TEs) in genomic sequences.
"""

import os
import sys
import getopt
import exceptions
import time
import ConfigParser
import glob

if not "REPET_PATH" in os.environ.keys():
    print "ERROR: no environment variable REPET_PATH"
    sys.exit(1)
sys.path.append( os.environ["REPET_PATH"] )
import pyRepet.launcher.programLauncher
import pyRepet.sql.RepetJobMySQL
import pyRepet.launcher.Launcher
from repet_base.RmvPairAlignInChunkOverlaps import RmvPairAlignInChunkOverlaps
from pyRepetUnit.commons.checker.CheckerUtils import CheckerUtils
from ConfigParser import NoOptionError
from ConfigParser import NoSectionError
from repet_base.MapClusterLauncher import MapClusterLauncher
from repet_base.MafftClusterLauncher import MafftClusterLauncher
from pyRepetUnit.commons.checker.CheckerException import CheckerException
from ConfigParser import MissingSectionHeaderError
from pyRepetUnit.commons.seq.FastaUtils import FastaUtils

#------------------------------------------------------------------------------

def setup_env( config ):

    """
    Setup the required environment.

    @param config: configuration file
    @type config: file handling
    """
        
    os.environ["REPET_HOST"] = config.get("repet_env","repet_host")
    os.environ["REPET_USER"] = config.get("repet_env","repet_user")
    os.environ["REPET_PW"] = config.get("repet_env","repet_pw")
    os.environ["REPET_DB"] = config.get("repet_env","repet_db")
    os.environ["REPET_PORT"] = config.get("repet_env","repet_port")

#------------------------------------------------------------------------------

def help():

    """
    Give the list of the command-line options.
    """

    print
    print "usage: ",sys.argv[0],"[ options ]"
    print "options:"
    print "     -h: this help"
    print "     -P: project name (<=15 symbols, alphanumeric or underscore)"
    print "     -C: configuration file"
    print "     -S: step (1/2/3/4/5)"
    print "     -s: program to self-align (default=Blaster/Pals)"
    print "     -c: program to cluster HSPs (Grouper/Recon/Piler or GrpRec/GrpPil/RecPil/GrpRecPil)"
    print "     -m: program to build MSAs (default=Map/Mafft/Muscle/Clustalw/Tcoffee/Prank)"
    print "     -v: verbose (0/default=1)"
    print

#------------------------------------------------------------------------------

def main():

    """
    This program is a pipeline for the de novo identification of transposable elements (TEs) in genomic sequences.
    """

    global projectName
    projectName = ""
    global configFileName
    configFileName = ""
    step = ""
    smplAlign = "Blaster"
    clustHsp = ""
    multAlign = "Map"
    global verbose
    verbose = 1
    global configFileHandle
    
    try:
        opts, args = getopt.getopt(sys.argv[1:],"hP:C:S:s:c:m:v:")
    except getopt.GetoptError, err:
        print str(err)
        help()
        sys.exit(1)
    for o, a in opts:
        if o == "-h":
            help()
            sys.exit(0)
        elif o == "-P":
            projectName = a
        elif o == "-C":
            configFileName = a
        elif o == "-S":
            step = a
        elif o == "-s":
            smplAlign = a
        elif o == "-c":
            clustHsp = a
        elif o == "-m":
            multAlign = a
        elif o == "-v":
            verbose = int(a)

    if projectName == "" or configFileName == "" or step == "":
        print "ERROR: missing compulsory options"
        help()
        sys.exit(1)

    if not CheckerUtils.isMax15Char(projectName):
        print "ERROR: project name must have 15 character max"
        help()
        sys.exit(1)
        
    if not CheckerUtils.isCharAlphanumOrUnderscore(projectName):
        print "ERROR: project name must contain only alphanumeric or underscore character"
        help()
        sys.exit(1)

    global config
    configFileHandle = open(configFileName)
    config = ConfigParser.ConfigParser()
    
    try :
        config.readfp( configFileHandle )
    except MissingSectionHeaderError:
        print "ERROR: config file " + configFileName + " must begin with a section header "  
        sys.exit(1)
    
    isOptionErrorRaised = False
    sectionName = "repet_env"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName)
    except NoSectionError:
        print "ERROR: the section " + sectionName + " must be in your config file : " + configFileName
        sys.exit(1)

    optionName = "repet_version"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, optionName) 
    except NoOptionError:
        print "ERROR: the option " +  optionName +" must be define in " + sectionName +" in your config file : " + configFileName
        isOptionErrorRaised = True 
    optionName = "repet_host"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, optionName) 
    except NoOptionError:
        print "ERROR: the option " +  optionName +" must be define in " + sectionName +" in your config file : " + configFileName
        isOptionErrorRaised = True 
            
    optionName = "repet_user"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, optionName) 
    except NoOptionError:
        print "ERROR: the option " +  optionName +" must be define in " + sectionName +" in your config file : " + configFileName
        isOptionErrorRaised = True 
            
    optionName = "repet_pw"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, optionName) 
    except NoOptionError:
        print "ERROR: the option " +  optionName +" must be define in " + sectionName +" in your config file : " + configFileName
        isOptionErrorRaised = True 
            
    optionName = "repet_db"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, optionName) 
    except NoOptionError:
        print "ERROR: the option " +  optionName +" must be define in " + sectionName +" in your config file : " + configFileName
        isOptionErrorRaised = True 
            
    optionName = "repet_port"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, optionName) 
    except NoOptionError:
        print "ERROR: the option " +  optionName +" must be define in " + sectionName +" in your config file : " + configFileName
        isOptionErrorRaised = True 
        
    if isOptionErrorRaised:
        sys.exit(1)

        
    setup_env( config )
    
    changeLogFileName = os.environ["REPET_PATH"] + "/CHANGELOG" 
    changeLogFileHandle = open(changeLogFileName, "r")
    
    try:
        CheckerUtils.checkConfigVersion(changeLogFileHandle, config)
    except CheckerException, e:
        print e.msg
        sys.exit(1)
    
    sectionName = "project"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName)
    except NoSectionError:
        print "ERROR: the section " + sectionName + " must be in your config file : " + configFileName
        sys.exit(1)
    
    isOptionErrorRaised = False
    optionName = "project_name"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, optionName) 
    except NoOptionError:
        print "ERROR: the option " +  optionName +" must be define in " + sectionName +" in your config file : " + configFileName
        isOptionErrorRaised = True    

    optionName = "project_dir"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, optionName) 
    except NoOptionError:
        print "ERROR: the option " +  optionName +" must be define in " + sectionName +" in your config file : " + configFileName
        isOptionErrorRaised = True    

    if isOptionErrorRaised :
        sys.exit(1)


    if verbose > 0:
        print "\nSTART TEdenovo.py (%s)" % ( time.strftime("%Y-%m-%d %H:%M:%S") )
        print "version %s" % ( config.get("repet_env","repet_version") )
        sys.stdout.flush()

    if config.get( "project", "project_name" ) != projectName:
        print "ERROR: project name different between configuration file and command-line option"
        sys.exit(1)
    if verbose > 0:
        print "project name = %s" % ( projectName ); sys.stdout.flush()

    global projectDir
    projectDir = config.get( "project", "project_dir" )
    if projectDir[-1] == "/":
        projectDir = projectDir[:-1]
    if os.path.exists( projectDir ):
        if verbose > 0:
            print "project directory = %s" % ( projectDir ); sys.stdout.flush()
        os.chdir( projectDir )
    else:
        print "ERROR: project directory %s doesn't exist" % ( projectDir )
        sys.exit(1)

    global pL
    pL = pyRepet.launcher.programLauncher.programLauncher()

    if step == "1":
        selfAlign( smplAlign )

    if step == "2":
        clusterHsp( smplAlign, clustHsp )

    if step == "3":
        buildConsensus( smplAlign, clustHsp, multAlign )

    if step == "4":
        detectFeatures( smplAlign, clustHsp, multAlign )

    if step == "5":
        classifConsensus( smplAlign, clustHsp, multAlign )

    if verbose > 0:
        print "version %s" % ( config.get("repet_env","repet_version") )
        print "END TEdenovo.py (%s)\n" % ( time.strftime("%Y-%m-%d %H:%M:%S") )
        sys.stdout.flush()

    return 0

#------------------------------------------------------------------------------

def selfAlign( smplAlign ):
    """
    step 1: self-alignment of the input data bank

    @param smplAlign: name of the alignment program (Blaster or Pals)
    @type smplAlign: string
    """

    if verbose > 0:
        print "beginning of step 1"
        print "self-alignment with %s" % ( smplAlign )
        sys.stdout.flush()

    if os.path.exists( projectName + "_" + smplAlign ):
        print "ERROR: directory " + projectName + "_" + smplAlign + "/ already exists"
        sys.exit(1)

    sectionName = "self_align"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName)
    except NoSectionError:
        print "ERROR: the section " + sectionName + " must be in your config file : " + configFileName
        sys.exit(1)

    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "blast")
    except NoOptionError:
        print "ERROR: the option blast must be define in " + sectionName +" in your config file : " + configFileName
        sys.exit(1)
    if config.get(sectionName,"blast") == "wu":
        if not (CheckerUtils.isExecutableInUserPath("wu-blastall") or CheckerUtils.isExecutableInUserPath("wu-formatdb")):
            print "ERROR: wu-blastall and wu-formatdb must be in your path"
            sys.exit(1)
    else:
        if not (CheckerUtils.isExecutableInUserPath("blastall") or CheckerUtils.isExecutableInUserPath("formatdb")):
            print "ERROR: ncbi blastall and formatdb must be in your path"
            sys.exit(1)
    if smplAlign == "Blaster":
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "Evalue") 
        except NoOptionError:
            print "ERROR: the option Evalue must be define in " + sectionName +" in your config file : " + configFileName
            sys.exit(1)
    if smplAlign == "Pals" and not CheckerUtils.isExecutableInUserPath("pals"):
        print "ERROR: 'pals' must be in your path"
        sys.exit(1)
    if smplAlign == "Yass" and not CheckerUtils.isExecutableInUserPath("yass"):
        print "ERROR: 'yass' must be in your path"
        sys.exit(1)
    if smplAlign == "Blat" and not CheckerUtils.isExecutableInUserPath("blat"):
        print "ERROR: 'blat' must be in your path"
        sys.exit(1)

    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "cluster") 
    except NoOptionError:
        print "ERROR: the option cluster must be define in " + sectionName +" in your config file : " + configFileName
        sys.exit(1)
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "clean") 
    except NoOptionError:
        print "ERROR: the option clean must be define in " + sectionName +" in your config file : " + configFileName
        sys.exit(1) 
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "length") 
    except NoOptionError:
        print "ERROR: the option length must be define in " + sectionName +" in your config file : " + configFileName
        sys.exit(1)
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "identity") 
    except NoOptionError:
        print "ERROR: the option identity must be define in " + sectionName +" in your config file : " + configFileName
        sys.exit(1)

    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "filter_HSP") 
    except NoOptionError:
        print "ERROR: the option filter_HSP must be define in " + sectionName +" in your config file : " + configFileName
        sys.exit(1)      
    if config.get(sectionName,"filter_HSP") == "yes":
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "min_Evalue") 
        except NoOptionError:
            print "ERROR: the option min_Evalue must be define in " + sectionName +" in your config file : " + configFileName
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "min_identity") 
        except NoOptionError:
            print "ERROR: the option min_identity must be define in " + sectionName +" in your config file : " + configFileName
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "min_length") 
        except NoOptionError:
            print "ERROR: the option min_length must be define in " + sectionName +" in your config file : " + configFileName
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "max_length") 
        except NoOptionError:
            print "ERROR: the option max_length must be define in " + sectionName +" in your config file : " + configFileName
            sys.exit(1)

    os.mkdir( projectName + "_" + smplAlign )
    os.chdir( projectName + "_" + smplAlign )
    os.system( "ln -s ../" + projectName + ".fa ." )


    # run in parallel (on chunks)
    if os.environ[ "REPET_QUEUE" ] != "" and config.get(sectionName,"cluster") == "yes":
        
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "chunk_length") 
        except NoOptionError:
            print "ERROR: the option chunk_length must be define in " + sectionName +" in your config file : " + configFileName
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "chunk_overlap") 
        except NoOptionError:
            print "ERROR: the option chunk_overlap must be define in " + sectionName +" in your config file : " + configFileName
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "nb_seq_per_batch") 
        except NoOptionError:
            print "ERROR: the option nb_seq_per_batch must be define in " + sectionName +" in your config file : " + configFileName
            sys.exit(1)       
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "queue") 
        except NoOptionError:
            print "ERROR: the option queue must be define in " + sectionName +" in your config file : " + configFileName
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "tmpDir") 
        except NoOptionError:
            print "ERROR: the option tmpDir must be define in " + sectionName +" in your config file : " + configFileName
            sys.exit(1)  
                                      
        if os.environ[ "REPET_QUEUE" ] == "SGE":
            if not CheckerUtils.isQueueNameValid(config.get(sectionName,"queue")):
                print "ERROR: the queue name specified in self_align section in your configuration file is not valid"
                sys.exit(1)
        
        # build the chunks
        prg = os.environ["REPET_PATH"] + "/bin/dbChunks.py"
        cmd = prg
        cmd += " -i %s.fa" % ( projectName )
        cmd += " -l %s" % ( config.get(sectionName,"chunk_length") )
        cmd += " -o %s" % ( config.get(sectionName,"chunk_overlap") )
        cmd += " -w 0"
        cmd += " -O %s_chunks" % ( projectName )
        cmd += " -c"
        cmd += " -v %i" % ( verbose - 1 )
        if config.get(sectionName,"clean") == "yes":
            cmd += " -c"
        pL.launch( prg, cmd )

        # save the chunks into batch files
        prg = os.environ["REPET_PATH"] + "/bin/dbSplit.py"
        cmd = prg
        cmd += " -i %s_chunks.fa" % ( projectName )
        cmd += " -n %s" % ( config.get(sectionName,"nb_seq_per_batch") )
        cmd += " -d"
        cmd += " -v %i" % ( verbose - 1 )
        pL.launch( prg, cmd )

        os.system( "cp ../%s ." % ( configFileName ) )


        # if the program is Blaster
        if smplAlign == "Blaster":

            # prepare the subject data bank
            pL.reset( projectName + "_chunks.fa" )
            pL.launchBlaster( prepareOnly="yes", outPrefix=projectName+"_chunks.fa-BLRaPrepare" )

            # launch Blaster on all the chunks in parallel
            prg = os.environ["REPET_PATH"] + "/bin/srptBlasterMatcher.py"
            cmd = prg
            cmd += " -g %s_TEdenovo_%s" % ( projectName, smplAlign )
            cmd += " -q %s/%s_%s/batches" % ( projectDir, projectName, smplAlign )
            cmd += " -s " + projectDir + "/" + projectName + "_" + smplAlign + "/" + projectName + "_chunks.fa"
            cmd += " -Q \"%s\"" % ( config.get(sectionName,"queue") )
            cmd += " -m 1"   # launch Blaster only
            cmd += " -B \""
            cmd += " -a"
            cmd += " -E " + config.get(sectionName,"Evalue")
            cmd += " -L " + config.get(sectionName,"length")
            cmd += " -I " + config.get(sectionName,"identity")
            if config.get(sectionName,"blast") == "wu":
                cmd += " -W"
            cmd += "\""
            if config.get(sectionName,"tmpDir" ) != "":
                cmd += " -d %s" % ( config.get(sectionName,"tmpDir") )
            cmd += " -Z align"
            cmd += " -C %s" % ( configFileName )
            if config.get(sectionName,"clean") == "yes":
                cmd += " -c"
            cmd += " -v %i" % ( verbose - 1 )
            log = os.system( cmd )
            if log != 0:
                print "ERROR: %s returned %i" % ( prg, log )
                sys.exit(1)

            log = os.system( "mv %s_TEdenovo_%s.align %s.align" % ( projectName, smplAlign, projectName ) )
            if log != 0:
                print "ERROR while renaming the 'align' file"
                sys.exit(1)

            # clean 'cut' files
            if config.get(sectionName,"clean") == "yes":
                os.system( "rm -f formatdb.log %s_chunks.fa_cut*" % ( projectName ) )


        # if the program is Pals
        elif smplAlign == "Pals":
            
                # launch Pals on all the chunks in parallel
                prg = os.environ["REPET_PATH"] + "/bin/srptPals.py"
                cmd = prg
                cmd += " -t jobs"
                cmd += " -g %s_TEdenovo_%s" % ( projectName, smplAlign )
                cmd += " -q %s/%s_%s/batches" % ( projectDir, projectName, smplAlign )
                cmd += " -s " + projectDir + "/" + projectName + "_" + smplAlign + "/" + projectName + "_chunks.fa"
                cmd += " -Q \"%s\"" % ( config.get(sectionName,"queue") )
                if config.get(sectionName,"tmpDir" ) != "":
                    cmd += " -d %s" % ( config.get(sectionName,"tmpDir") )
                cmd += " -C %s" % ( configFileName )
                cmd += " -P \""
                cmd += "-length " + config.get(sectionName,"length")  #400
                cmd += " -pctid " + config.get(sectionName,"identity")  #94
                cmd += "\""
                cmd += " -Z"
                cmd += " -a"
                if config.get(sectionName,"clean") == "yes":
                    cmd += " -c"
                log = os.system( cmd )
                if log != 0:
                    print "ERROR: %s returned %i" % ( prg, log )
                    sys.exit(1)
    
                os.system( "mv %s_TEdenovo_%s.align %s.align" % ( projectName, smplAlign, projectName ) )

        # if the program is Yass or Blat
        elif smplAlign in [ "Yass", "Blat" ]:
            if smplAlign == "Yass":
                prg = os.environ["REPET_PATH"] + "/bin/YassClusterLauncher.py"
            elif smplAlign == "Blat":
                prg = os.environ["REPET_PATH"] + "/bin/BlatClusterLauncher.py"
            cmd = prg
            cmd += " -i %s/%s_%s/batches" % ( projectDir, projectName, smplAlign )
            cmd += " -Q \"%s\"" % ( config.get(sectionName,"queue") )
            cmd += " -g %s_TEdenovo_%s" % ( projectName, smplAlign )
            cmd += " -C %s" % ( configFileName )
            if config.get(sectionName,"tmpDir") != "":
                cmd += " -d %s" % ( config.get(sectionName,"tmpDir") )
            cmd += " -j jobs"
            cmd += " -s %s/%s_%s/%s_chunks.fa" % ( projectDir, projectName, smplAlign, projectName )
            if smplAlign == "Yass":
                cmd += " -p \"-E %s\"" % ( config.get(sectionName,"Evalue") )
            elif smplAlign == "Blat":
                cmd += " -p \"-minIdentity=%s\"" % ( config.get(sectionName,"identity") )
            cmd += " -Z"
            cmd += " -A"
            if config.get(sectionName,"clean") == "yes":
                cmd += " -c"
            cmd += " -v %i" % ( verbose - 1 )
            log = os.system( cmd )
            if log != 0:
                print "ERROR: %s returned %i" % ( prg, log )
                sys.exit(1)
                
            os.system( "mv %s_TEdenovo_%s.align %s.align" % ( projectName, smplAlign, projectName ) )

        # remove the chunk overlaps
        rpaico = RmvPairAlignInChunkOverlaps()
        rpaico.setInputFileName( "%s.align" % ( projectName ) )
        rpaico.setChunkLength( config.get(sectionName,"chunk_length") )
        rpaico.setChunkOverlap( config.get(sectionName,"chunk_overlap") )
        rpaico.setMargin( 10 )
        rpaico.setOutputFileName( "%s.align.not_over" % ( projectName ) )
        rpaico.setVerbosityLevel( verbose - 1 )
        rpaico.run()

    # or run not in parallel
    else:
        if smplAlign == "Blaster":
            prg = os.environ["REPET_PATH"] + "/bin/blaster"
            cmd = prg
            cmd += " -q %s.fa" % ( projectName )
            cmd += " -s %s.fa" % ( projectName )
            cmd += " -a"
            cmd += " -E %s" % ( config.get(sectionName,"Evalue") )
            cmd += " -L %s" % ( config.get(sectionName,"length") )
            cmd += " -I %s" % ( config.get(sectionName,"identity") )
            if config.get(sectionName,"blast") == "wu":
                cmd += " -W"
            cmd += " -B %s" % ( projectName )
            cmd += " -v %i" % ( verbose )
            print cmd; sys.stdout.flush()
            pL.launch( prg, cmd )
            if config.get(sectionName,"clean") == "yes":
                os.system( "rm -f formatdb.log %s.raw %s.seq_treated" % ( projectName, projectName ) )
        elif smplAlign == "Pals":
            prg = "pals"
            cmd = prg
            cmd += " -self %s.fa" % ( projectName )
            cmd += " -out %s.gff" % ( projectName )
            cmd += " -length %s" % ( config.get(sectionName,"length") )
            cmd += " -pctid %s" % ( config.get(sectionName,"identity") )
            print cmd; sys.stdout.flush()
            pL.launch( prg, cmd )
            prg = os.environ["REPET_PATH"] + "/bin/pals2align.py"
            cmd = prg
            cmd += " -i %s.gff" % ( projectName )
            cmd += " -o %s.align" % ( projectName )
            log = os.system( cmd )
            if log != 0:
                print "ERROR: %s returned %i" % ( prg, log )
                sys.exit(1)
            os.remove( "%s.gff" % ( projectName ) )
        elif smplAlign == "Yass":
            prg = "yass"
            cmd = prg
            cmd += " -d 2"
            cmd += " -E %s" % ( config.get(sectionName,"Evalue") )
            cmd += " -o %s.blast" % ( projectName )
            cmd += " %s.fa" % ( projectName )
            cmd += " %s.fa" % ( projectName )
            print cmd; sys.stdout.flush()
            pL.launch( prg, cmd )
            prg = os.environ["REPET_PATH"] + "/bin/blast2align.py"
            cmd = prg
            cmd += " -i %s.blast" % ( projectName )
            cmd += " -o %s.align" % ( projectName )
            log = os.system( cmd )
            if log != 0:
                print "ERROR: %s returned %i" % ( prg, log )
                sys.exit(1)
            os.remove( "%s.blast" % ( projectName ) )
        elif smplAlign == "Blat":
            prg = "blat"
            cmd = prg
            cmd += " %s.fa" % ( projectName )
            cmd += " %s.fa" % ( projectName )
            cmd += " -minIdentity=%s" % ( config.get(sectionName,"identity") )
            cmd += " -out=blast8"
            cmd += " %s.blast" % ( projectName )
            print cmd; sys.stdout.flush()
            pL.launch( prg, cmd )
            prg = os.environ["REPET_PATH"] + "/bin/blast2align.py"
            cmd = prg
            cmd += " -i %s.blast" % ( projectName )
            cmd += " -o %s.align" % ( projectName )
            log = os.system( cmd )
            if log != 0:
                print "ERROR: %s returned %i" % ( prg, log )
                sys.exit(1)
            os.remove( "%s.blast" % ( projectName ) )
        else:
            print "ERROR: %s not implemented" % ( smplAlign )
            sys.exit(1)


    # filter the HSPs
    if config.get(sectionName,"filter_HSP") == "yes":
        prg = ""
        if os.path.exists( os.environ["REPET_PATH"] + "/bin/filterAlign" ):
            prg = os.environ["REPET_PATH"] + "/bin/filterAlign"
        else:
            prg = os.environ["REPET_PATH"] + "/bin/filterOutBlaster.py"
        cmd = prg
        cmd += " -i "
        if config.get(sectionName,"cluster") == "yes":
            cmd += "%s.align.not_over" % ( projectName )
        if config.get(sectionName,"cluster") == "no":
            cmd += "%s.align" % ( projectName )
        cmd += " -E %s" % ( config.get(sectionName,"min_Evalue") )
        cmd += " -I %s" % ( config.get(sectionName,"min_identity") )
        cmd += " -l %s" % ( config.get(sectionName,"min_length") )
        cmd += " -L %s" % ( config.get(sectionName,"max_length") )
        cmd += " -v %i" % ( verbose - 1 )
        pL.launch( prg, cmd )

    # clean
    if config.get(sectionName,"clean") == "yes":

        # for each job, remove the file containing the HSPs
        cmd = "find . -name \"batch_*.align\" -exec rm {} \;"
        log = os.system( cmd )
        if log != 0:
            print "ERROR while removing the 'align' files from the jobs"
            sys.exit(1)

        # remove the files "_cut"
        cmd = "find . -name \"%s*_cut*\" -exec rm {} \;" % ( projectName )
        log = os.system( cmd )
        if log != 0:
            print "ERROR while removing the '_cut' files"
            sys.exit(1)

        # remove the files "Nstretch.map"
        cmd = "find . -name \"*Nstretch.map\" -exec rm {} \;"
        log = os.system( cmd )
        if log != 0:
            print "ERROR while removing the 'Nstretch.map' files"
            sys.exit(1)

        # remove the temporary file(s) with the HSPs
        if config.get(sectionName,"cluster") == "yes" and config.get(sectionName,"filter_HSP") == "yes":
            cmd = "rm -f %s.align %s.align.not_over" % ( projectName, projectName )
            log = os.system( cmd )
            if log != 0:
                print "ERROR while removing an intermediary file"
                sys.exit(1)
        if config.get(sectionName,"cluster") == "no" and config.get(sectionName,"filter_HSP") == "yes":
            cmd = "rm -f %s.align" % ( projectName )
            log = os.system( cmd )
            if log != 0:
                print "ERROR while removing an intermediary file"
                sys.exit(1)

        # remove the genomic sequences (symlink and batches)
        if config.get(sectionName,"cluster") == "yes":
            os.system( "rm -rf %s.fa batches" % ( projectName ) )

    os.chdir( ".." )

    if verbose > 0:
        print "step 1 finished successfully"; sys.stdout.flush()

#------------------------------------------------------------------------------

def clusterHsp( smplAlign, clustHsp ):
    """
    step 2: cluster the resulting HSPs (high-scoring segment pairs)

    @param smplAlign: name of the alignment program (Blaster or Pals)
    @type smplAlign: string

    @param clustHsp: name of the clustering program (Grouper, Recon or Piler)
    @type clustHsp: string
    """

    if verbose > 0:
        print "beginning of step 2"
        print "self-alignment with %s" % ( smplAlign )
        print "clustering with %s" % ( clustHsp )
        sys.stdout.flush()

    sectionName = "self_align"
    sectionNameClusterHSP = "cluster_HSPs"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionNameClusterHSP)
    except NoSectionError:
        print "ERROR: " + sectionNameClusterHSP + " must be in your config file : " + configFileName
        sys.exit(1)

    if clustHsp == "Grouper" or clustHsp == "Recon" or clustHsp == "Piler":
        if os.path.exists( projectName + "_" + smplAlign + "_" + clustHsp ):
            print "ERROR: directory " + projectName + "_" + smplAlign + "_" + clustHsp +"/ already exists"
            sys.exit(1)
            
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionNameClusterHSP, "minNbSeqPerGroup") 
        except NoOptionError:
            print "ERROR: the option minNbSeqPerGroup must be define in " + sectionNameClusterHSP +" in your config file : " + configFileName
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionNameClusterHSP, "nbLongestSeqPerGroup") 
        except NoOptionError:
            print "ERROR: the option nbLongestSeqPerGroup must be define in " + sectionNameClusterHSP +" in your config file : " + configFileName
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "cluster") 
        except NoOptionError:
            print "ERROR: the option cluster must be define in " + sectionName +" in your config file : " + configFileName
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "filter_HSP") 
        except NoOptionError:
            print "ERROR: the option filter_HSP must be define in " + sectionName +" in your config file : " + configFileName
            sys.exit(1)

        os.mkdir( projectName + "_" + smplAlign + "_" + clustHsp )
        os.chdir( projectName + "_" + smplAlign + "_" + clustHsp )

        prefix = projectName + "_" + smplAlign + "_" + clustHsp + "_"
        prefix += config.get("cluster_HSPs","minNbSeqPerGroup") + "elem_"
        prefix += config.get("cluster_HSPs","nbLongestSeqPerGroup") + "seq"

        alignFileName = projectName + ".align"
        if config.get(sectionName,"cluster") == "yes":
            alignFileName += ".not_over"
        if config.get(sectionName,"filter_HSP") == "yes":
            alignFileName += ".filtered"
        pathToAlignFile = "%s/%s_%s/%s" % ( projectDir, projectName, smplAlign, alignFileName )
        if os.path.exists( "%s.onChr" % ( pathToAlignFile ) ):
            alignFileName += ".onChr"
            pathToAlignFile += ".onChr"

        bankName = projectName
        if config.get(sectionName,"cluster") == "yes":
            bankName += "_chunks.fa"
        else:
            bankName += ".fa"
        if "onChr" in alignFileName:
            bankName = "%s.fa" % ( projectName )
        pathToFasta = "%s/%s_%s/%s" % ( projectDir, projectName, smplAlign, bankName )

        os.system( "ln -s " + pathToAlignFile + " " + alignFileName )
        os.system( "ln -s %s ." % ( pathToFasta ) )

        #----------------------------------------------------------------------

        lCmds = []
        minSeq = config.get("cluster_HSPs","minNbSeqPerGroup")
        maxSeq = config.get("cluster_HSPs","nbLongestSeqPerGroup")

        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionNameClusterHSP, "cluster") 
        except NoOptionError:
            print "ERROR: the option cluster must be define in " + sectionNameClusterHSP +" in your config file : " + configFileName
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionNameClusterHSP, "clean") 
        except NoOptionError:
            print "ERROR: the option clean must be define in " + sectionNameClusterHSP +" in your config file : " + configFileName
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionNameClusterHSP, "maxSeqLength") 
        except NoOptionError:
            print "ERROR: the option maxSeqLength must be define in " + sectionNameClusterHSP +" in your config file : " + configFileName
            sys.exit(1)
                
        if config.get("cluster_HSPs","cluster") == "yes":
            try:
                CheckerUtils.checkOptionInSectionInConfigFile(config, sectionNameClusterHSP, "queue") 
            except NoOptionError:
                print "ERROR: the option queue must be define in " + sectionNameClusterHSP +" in your config file : " + configFileName
                sys.exit(1)
            try:
                CheckerUtils.checkOptionInSectionInConfigFile(config, sectionNameClusterHSP, "tmpDir") 
            except NoOptionError:
                print "ERROR: the option tmpDir must be define in " + sectionNameClusterHSP +" in your config file : " + configFileName
                sys.exit(1)
            os.system( "ln -s ../%s ." % ( configFileName ) )
            jobdb = pyRepet.sql.RepetJobMySQL.RepetJob( cfgFileName=configFileName )
            queue = config.get("cluster_HSPs","queue")
            if os.environ[ "REPET_QUEUE" ] == "SGE":
                if not CheckerUtils.isQueueNameValid(queue):
                    print "ERROR: the queue name specified in cluster_HSPs section in your configuration file is not valid"
                    sys.exit(1)
            groupid = "%s_%s_%s" % ( projectName, smplAlign, clustHsp )
            cDir = os.getcwd()
            if config.get("cluster_HSPs","tmpDir" ) != "":
                tmpDir = config.get("cluster_HSPs","tmpDir")
            else:
                tmpDir = cDir
            acronym = "%s_%s" % ( clustHsp, os.getpid() )
            cL = pyRepet.launcher.Launcher.Launcher( jobdb, os.getcwd(), "", "", cDir, tmpDir, "jobs", queue, groupid, acronym )
            cL.beginRun()
            cmd_start = ""
            cmd_start += "os.system( \"ln -s " + cDir + "/" + bankName + " " + bankName + "\" )\n"
            cmd_start += "os.system( \"ln -s " + cDir + "/" + alignFileName + " " + alignFileName + "\" )\n"
            cmd_finish = ""
            cmd_finish += "if os.path.exists( \"" + tmpDir + "/" + bankName + "\" ):\n"
            cmd_finish += "\tos.system( \"rm -f " + tmpDir + "/" + bankName + "\" )\n"
            cmd_finish += "if os.path.exists( \"" + tmpDir + "/" + alignFileName + "\" ):\n"
            cmd_finish += "\tos.system( \"rm -f " + tmpDir + "/" + alignFileName + "\" )\n"

        if clustHsp == "Grouper":
            try:
                CheckerUtils.checkOptionInSectionInConfigFile(config, sectionNameClusterHSP, "Grouper_coverage") 
            except NoOptionError:
                print "ERROR: the option Grouper_coverage must be define in " + sectionNameClusterHSP +" in your config file : " + configFileName
                sys.exit(1)
            try:
                CheckerUtils.checkOptionInSectionInConfigFile(config, sectionNameClusterHSP, "Grouper_join") 
            except NoOptionError:
                print "ERROR: the option Grouper_join must be define in " + sectionNameClusterHSP +" in your config file : " + configFileName
                sys.exit(1)
            try:
                CheckerUtils.checkOptionInSectionInConfigFile(config, sectionNameClusterHSP, "Grouper_include") 
            except NoOptionError:
                print "ERROR: the option Grouper_include must be define in " + sectionNameClusterHSP +" in your config file : " + configFileName
                sys.exit(1)
            try:
                CheckerUtils.checkOptionInSectionInConfigFile(config, sectionNameClusterHSP, "Grouper_maxJoinLength") 
            except NoOptionError:
                print "ERROR: the option Grouper_maxJoinLength must be define in " + sectionNameClusterHSP +" in your config file : " + configFileName
                sys.exit(1)
            coverage = config.get("cluster_HSPs","Grouper_coverage")
            prg = os.environ["REPET_PATH"] + "/bin/grouper"
            cmd = prg
            cmd += " -q %s" % ( bankName )
            cmd += " -s %s" % ( bankName )
            cmd += " -m %s" % ( alignFileName )
            if config.get("cluster_HSPs","Grouper_join") == "yes":
                cmd += " -j"
            cmd += " -C %s" % ( coverage )
            cmd += " -Z %s" % ( config.get("cluster_HSPs","minNbSeqPerGroup") )
            cmd += " -X %s" % ( config.get("cluster_HSPs","Grouper_include") )
            cmd += " -G -1"
            cmd += " -v %i" % ( verbose )
            lCmds.append( cmd )
            if config.get("cluster_HSPs","cluster") == "yes":
                cmd_finish += "if not os.path.exists( \"" + cDir + "/" + alignFileName + ".group.c" + coverage + ".fa \" ):\n"
                cmd_finish += "\tos.system( \"mv " + alignFileName + ".group.c" + coverage + ".fa " + cDir + "\" )\n"
                cmd_finish += "if not os.path.exists( \"" + cDir + "/" + alignFileName + ".group.c" + coverage + ".map \" ):\n"
                cmd_finish += "\tos.system( \"mv " + alignFileName + ".group.c" + coverage + ".map " + cDir + "\" )\n"
                cmd_finish += "if not os.path.exists( \"" + cDir + "/" + alignFileName + ".group.c" + coverage + ".param \" ):\n"
                cmd_finish += "\tos.system( \"mv " + alignFileName + ".group.c" + coverage + ".param " + cDir + "\" )\n"
                cmd_finish += "if not os.path.exists( \"" + cDir + "/" + alignFileName + ".group.c" + coverage + ".set \" ):\n"
                cmd_finish += "\tos.system( \"mv " + alignFileName + ".group.c" + coverage + ".set " + cDir + "\" )\n"
                cmd_finish += "if not os.path.exists( \"" + cDir + "/" + alignFileName + ".group.c" + coverage + ".txt \" ):\n"
                cmd_finish += "\tos.system( \"mv " + alignFileName + ".group.c" + coverage + ".txt " + cDir + "\" )\n"
            prg = os.environ["REPET_PATH"] + "/bin/filterOutGrouper.py"
            cmd = prg
            cmd += " -i %s.group.c%s.fa" % ( alignFileName, config.get("cluster_HSPs","Grouper_coverage") )
            cmd += " -m %s" % ( minSeq )
            cmd += " -M %s" % ( maxSeq )
            cmd += " -L %s" % ( config.get("cluster_HSPs","maxSeqLength" ) )
            cmd += " -J %s" % ( config.get("cluster_HSPs","Grouper_maxJoinLength" ) )
            if config.get(sectionName,"cluster") == "yes" and "onChr" not in alignFileName:
                try:
                    CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "chunk_overlap") 
                except NoOptionError:
                    print "ERROR: the option chunk_overlap must be define in " + sectionName +" in your config file : " + configFileName
                    sys.exit(1)
                cmd += " -O %s" % ( config.get(sectionName,"chunk_overlap") )
            else:
                cmd += " -O -1"
            cmd += " -o %s_%s_%s_%selem_%sseq.fa" % ( projectName, smplAlign, clustHsp, minSeq, maxSeq )
            cmd += " -v %i" % ( verbose - 1 )
            lCmds.append( cmd )
            if config.get("cluster_HSPs","cluster") == "yes":
                cmd_finish += "if not os.path.exists( \"%s/%s_%s_%s_%selem_%sseq.fa\" ):\n" % ( cDir, projectName, smplAlign, clustHsp, minSeq, maxSeq )
                cmd_finish += "\tos.system( \"mv %s_%s_%s_%selem_%sseq.fa %s\" )\n" % ( projectName, smplAlign, clustHsp, minSeq, maxSeq, cDir )
                cmd_finish += "if not os.path.exists( \"%s/%s.group.c%s.fa_filtered.log\" ):\n" % ( cDir, alignFileName, config.get("cluster_HSPs","Grouper_coverage") )
                cmd_finish += "\tos.system( \"mv %s.group.c%s.fa_filtered.log %s\" )\n" % ( alignFileName, config.get("cluster_HSPs","Grouper_coverage"), cDir )

        elif clustHsp == "Recon":
            if not CheckerUtils.isExecutableInUserPath("recon.pl") :
                print "ERROR: recon.pl must be in your path"
                sys.exit(1)
            prg = os.environ["REPET_PATH"] + "/bin/launchRecon.py"
            cmd = prg
            cmd += " -p %s_%s" % ( projectName, smplAlign )
            cmd += " -f %s" % ( bankName )
            cmd += " -a %s"  % ( alignFileName )
            if config.get("cluster_HSPs","clean") == "yes" or config.get("cluster_HSPs","cluster") == "yes":
                cmd += " -c"
            cmd += " -v %i" % ( verbose - 1 )
            lCmds.append( cmd )
            if config.get("cluster_HSPs","cluster") == "yes":
                cmd_finish += "if not os.path.exists( \"" + cDir + "/" + projectName + "_" + smplAlign + "_Recon.map \" ):\n"
                cmd_finish += "\tos.system( \"mv " + projectName + "_" + smplAlign + "_Recon.map " + cDir + "\" )\n"
                cmd_finish += "if not os.path.exists( \"" + cDir + "/" + projectName + "_" + smplAlign + "_Recon.fa \" ):\n"
                cmd_finish += "\tos.system( \"mv " + projectName + "_" + smplAlign + "_Recon.fa " + cDir + "\" )\n"
                cmd_finish += "if not os.path.exists( \"" + cDir + "/" + projectName + "_" + smplAlign + "_MSP_file \" ):\n"
                cmd_finish += "\tos.system( \"mv " + projectName + "_" + smplAlign + "_MSP_file " + cDir + "\" )\n"
                cmd_finish += "if not os.path.exists( \"" + cDir + "/" + projectName + "_" + smplAlign + "_seq_list \" ):\n"
                cmd_finish += "\tos.system( \"mv " + projectName + "_" + smplAlign + "_seq_list " + cDir + "\" )\n"
            prg = os.environ["REPET_PATH"] + "/bin/filterSeqClusters.py"
            cmd = prg
            cmd += " -i %s_%s_%s.map" % ( projectName, smplAlign, clustHsp )
            cmd += " -c %s" % ( clustHsp )
            cmd += " -m %s" % ( minSeq )
            cmd += " -M %s" % ( maxSeq )
            cmd += " -L %s" % ( config.get("cluster_HSPs","maxSeqLength" ) )
            cmd += " -d"
            cmd += " -v %i" % ( verbose - 1 )
            lCmds.append( cmd )
            prg = os.environ["REPET_PATH"] + "/bin/map2db"
            cmd = prg
            cmd += " %s_%s_%s.map.filtered-%s-%s" % ( projectName, smplAlign, clustHsp, minSeq, maxSeq )
            cmd += " %s" % ( bankName )
            lCmds.append( cmd )
            if config.get("cluster_HSPs","cluster") == "yes":
                cmd_finish += "if not os.path.exists( \"%s/%s_%s_%s.map.filtered-%s-%s\" ):\n" % ( cDir, projectName, smplAlign, clustHsp, minSeq, maxSeq )
                cmd_finish += "\tos.system( \"mv %s_%s_%s.map.filtered-%s-%s %s\" )\n" % ( projectName, smplAlign, clustHsp, minSeq, maxSeq, cDir )
                cmd_finish += "if not os.path.exists( \"%s/%s_%s_%s.map_filtered.log\" ):\n" % ( cDir,projectName, smplAlign, clustHsp )
                cmd_finish += "\tos.system( \"mv %s_%s_%s.map_filtered.log %s\" )\n" % ( projectName, smplAlign, clustHsp, cDir )
                cmd_finish += "if not os.path.exists( \"%s/%s_%s_%s.map.distribC\" ):\n" % ( cDir, projectName, smplAlign, clustHsp )
                cmd_finish += "\tos.system( \"mv %s_%s_%s.map.distribC %s\" )\n" % ( projectName, smplAlign, clustHsp, cDir )
                cmd_finish += "if not os.path.exists( \"%s/%s_%s_%s.map.filtered-%s-%s.flank_size0.fa\" ):\n" % ( cDir, projectName, smplAlign, clustHsp, minSeq, maxSeq )
                cmd_finish += "\tos.system( \"mv %s_%s_%s.map.filtered-%s-%s.flank_size0.fa %s\" )\n" % ( projectName, smplAlign, clustHsp, minSeq, maxSeq, cDir )

        elif clustHsp == "Piler":
            if not CheckerUtils.isExecutableInUserPath("piler") :
                print "ERROR: piler must be in your path"
                sys.exit(1)
            prg = os.environ["REPET_PATH"] + "/bin/launchPiler.py"
            cmd = prg
            cmd += " -p %s_%s" % ( projectName, smplAlign )
            cmd += " -f %s" % ( bankName )
            cmd += " -a %s"  % ( alignFileName )
            if config.get("cluster_HSPs","clean") == "yes" or config.get("cluster_HSPs","cluster") == "yes":
                cmd += " -c"
            cmd += " -v %i" % ( verbose - 1 )
            lCmds.append( cmd )
            if config.get("cluster_HSPs","cluster") == "yes":
                cmd_finish += "if not os.path.exists( \"" + cDir + "/" + projectName + "_" + smplAlign + "_Piler.map \" ):\n"
                cmd_finish += "\tos.system( \"mv " + projectName + "_" + smplAlign + "_Piler.map " + cDir + "\" )\n"
                cmd_finish += "if not os.path.exists( \"" + cDir + "/" + projectName + "_" + smplAlign + "_Piler.fa \" ):\n"
                cmd_finish += "\tos.system( \"mv " + projectName + "_" + smplAlign + "_Piler.fa " + cDir + "\" )\n"
                cmd_finish += "if not os.path.exists( \"" + cDir + "/" + alignFileName + ".gff \" ):\n"
                cmd_finish += "\tos.system( \"mv " + alignFileName + ".gff " + cDir + "\" )\n"
                cmd_finish += "if not os.path.exists( \"" + cDir + "/" + projectName + "_" + smplAlign + "_Piler-trs.gff \" ):\n"
                cmd_finish += "\tos.system( \"mv " + projectName + "_" + smplAlign + "_Piler-trs.gff " + cDir + "\" )\n"
            prg = os.environ["REPET_PATH"] + "/bin/filterSeqClusters.py"
            cmd = prg
            cmd += " -i %s_%s_%s.map" % ( projectName, smplAlign, clustHsp )
            cmd += " -c %s" % ( clustHsp )
            cmd += " -m %s" % ( minSeq )
            cmd += " -M %s" % ( maxSeq )
            cmd += " -L %s" % ( config.get("cluster_HSPs","maxSeqLength" ) )
            cmd += " -d"
            cmd += " -v %i" % ( verbose - 1 )
            lCmds.append( cmd )
            prg = os.environ["REPET_PATH"] + "/bin/map2db"
            cmd = prg
            cmd += " %s_%s_%s.map.filtered-%s-%s" % ( projectName, smplAlign, clustHsp, minSeq, maxSeq )
            cmd += " %s" % ( bankName )
            lCmds.append( cmd )
            if config.get("cluster_HSPs","cluster") == "yes":
                cmd_finish += "if not os.path.exists( \"%s/%s_%s_%s.map.filtered-%s-%s\" ):\n" % ( cDir, projectName, smplAlign, clustHsp, minSeq, maxSeq )
                cmd_finish += "\tos.system( \"mv %s_%s_%s.map.filtered-%s-%s %s\" )\n" % ( projectName, smplAlign, clustHsp, minSeq, maxSeq, cDir )
                cmd_finish += "if not os.path.exists( \"%s/%s_%s_%s.map_filtered.log\" ):\n" % ( cDir,projectName, smplAlign, clustHsp )
                cmd_finish += "\tos.system( \"mv %s_%s_%s.map_filtered.log %s\" )\n" % ( projectName, smplAlign, clustHsp, cDir )
                cmd_finish += "if not os.path.exists( \"%s/%s_%s_%s.map.distribC\" ):\n" % ( cDir, projectName, smplAlign, clustHsp )
                cmd_finish += "\tos.system( \"mv %s_%s_%s.map.distribC %s\" )\n" % ( projectName, smplAlign, clustHsp, cDir )
                cmd_finish += "if not os.path.exists( \"%s/%s_%s_%s.map.filtered-%s-%s.flank_size0.fa\" ):\n" % ( cDir, projectName, smplAlign, clustHsp, minSeq, maxSeq )
                cmd_finish += "\tos.system( \"mv %s_%s_%s.map.filtered-%s-%s.flank_size0.fa %s\" )\n" % ( projectName, smplAlign, clustHsp, minSeq, maxSeq, cDir )

        if config.get("cluster_HSPs","cluster") == "no":
            for c in lCmds:
                log = os.system( c )
                if log != 0:
                    print "ERROR with command '%s'" % ( c )
                    sys.exit(1)
        elif config.get("cluster_HSPs","cluster") == "yes":
            for c in lCmds:
                cmd_start += "log = os.system( \""
                cmd_start += c
                cmd_start += "\" )\n"
            cL.runSingleJob( cmd_start, cmd_finish )
            cL.endRun()
            if config.get("cluster_HSPs","clean") == "yes":
                cL.clean( "%s" % ( clustHsp ) )

        if clustHsp in [ "Recon", "Piler" ]:
            cmd = "ln -s"
            cmd += " %s_%s_%s.map.filtered-%s-%s.flank_size0.fa" % ( projectName, smplAlign, clustHsp, minSeq, maxSeq )
            cmd += " %s_%s_%s_%selem_%sseq.fa" % ( projectName, smplAlign, clustHsp, minSeq, maxSeq )
            log = os.system( cmd )
            if log != 0:
                print "ERROR with command '%s'" % ( cmd )
                sys.exit(1)

        os.chdir( ".." )

    #--------------------------------------------------------------------------

    elif clustHsp == "":
        print "ERROR: missing option -c"
        help()
        sys.exit(1)

    else:
        print "ERROR: unknown option -c %s" % ( clustHsp )
        sys.exit(1)

    if verbose > 0:
        print "step 2 finished successfully"; sys.stdout.flush()

#------------------------------------------------------------------------------

def buildConsensus( smplAlign, clustHsp, multAlign ):

    """
    step 3: build consensus from multiple alignments

    @param smplAlign: name of the alignment program (Blaster or Pals)
    @type smplAlign: string

    @param clustHsp: name of the clustering program (Grouper, Recon or Piler)
    @type clustHsp: string

    @param multAlign: name of the multiple alignment program MSA (MAP, Mafft, Muscle, Tcoffee, ClustalW or Prank)
    @type multAlign: string
    """

    if verbose > 0:
        print "beginning of step 3"
        print "self-alignment with %s" % ( smplAlign )
        print "clustering with %s" % ( clustHsp )
        print "multiple alignment with %s" % ( multAlign )
        sys.stdout.flush()

    isOptionErrorRaised = False

    sectionNameClusterHSPs = "cluster_HSPs"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionNameClusterHSPs) 
    except NoSectionError:
        print "ERROR: the section " + sectionNameClusterHSPs + " must be define in your config file : " + configFileName
        sys.exit(1)
 
    minNbSeqPerGroup = "minNbSeqPerGroup"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionNameClusterHSPs, minNbSeqPerGroup) 
    except NoOptionError:
        print "ERROR: the option " + minNbSeqPerGroup + " must be define in " + sectionNameClusterHSPs +" in your config file : " + configFileName
        isOptionErrorRaised = True
    
    nbLongestSeqPerGroup = "nbLongestSeqPerGroup"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionNameClusterHSPs, nbLongestSeqPerGroup) 
    except NoOptionError:
        print "ERROR: the option " + nbLongestSeqPerGroup + " must be define in " + sectionNameClusterHSPs +" in your config file : " + configFileName
        isOptionErrorRaised = True
   
    sectionNameBuildConsensus = "build_consensus" 
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionNameBuildConsensus) 
    except NoSectionError:
        print "ERROR: the section " + sectionNameBuildConsensus + " must be define in your config file : " + configFileName
        sys.exit(1)
   
    queue = "queue"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionNameBuildConsensus, queue) 
    except NoOptionError:
        print "ERROR: the option " + queue + " must be define in " + sectionNameBuildConsensus +" in your config file : " + configFileName
        isOptionErrorRaised = True
 
    tmpDir = "tmpDir"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionNameBuildConsensus, tmpDir) 
    except NoOptionError:
        print "ERROR: the option " + tmpDir + " must be define in " + sectionNameBuildConsensus +" in your config file : " + configFileName
        isOptionErrorRaised = True

    minBasesPerSite = "minBasesPerSite"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionNameBuildConsensus, minBasesPerSite) 
    except NoOptionError:
        print "ERROR: the option " + minBasesPerSite + " must be define in " + sectionNameBuildConsensus +" in your config file : " + configFileName
        isOptionErrorRaised = True

    clean = "clean"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionNameBuildConsensus, clean) 
    except NoOptionError:
        print "ERROR: the option " + clean + " must be define in " + sectionNameBuildConsensus +" in your config file : " + configFileName
        isOptionErrorRaised = True
    
    if isOptionErrorRaised:
        sys.exit(1)
    
    prefix1 = projectName + "_" + smplAlign + "_" + clustHsp
    prefix2 = prefix1
    prefix2 += "_" + config.get("cluster_HSPs","minNbSeqPerGroup") + "elem_"
    prefix2 += config.get("cluster_HSPs","nbLongestSeqPerGroup") + "seq"

    if clustHsp == "Grouper" or clustHsp == "Recon" or clustHsp == "Piler":

        if not os.path.exists( "%s/%s.fa" % ( prefix1, prefix2 ) ):
            print "ERROR: %s/%s.fa not found" % ( prefix1, prefix2 )
            sys.exit(1)

        if os.path.exists( "%s_%s" % ( prefix1, multAlign ) ):
            print "ERROR: directory %s_%s/ already exists" % ( prefix1, multAlign )
            sys.exit(1)

        if multAlign not in ["Map","MAP","Mafft","Clustalw","Muscle","Prank","Tcoffee"]:
            print "ERROR: %s not implememented" % ( multAlign )
            sys.exit(1)

        os.mkdir( "%s_%s" % ( prefix1, multAlign ) )
        os.chdir( "%s_%s" % ( prefix1, multAlign ) )

        prg = os.environ["REPET_PATH"] + "/bin/splitSeqPerCluster.py"
        cmd = prg
        cmd += " -i ../" + prefix1 + "/" + prefix2 + ".fa"
        cmd += " -c %s" % ( clustHsp )
        cmd += " -v %i" % ( verbose - 1 )
        pL.launch( prg, cmd )

        mapChunkFile = "%s/%s_%s/%s_chunks.map" % ( projectDir, projectName, smplAlign, projectName )
        if os.path.exists( mapChunkFile ):
            lFiles = glob.glob( "seqCluster*.fa" )
            f = open( lFiles[0] )
            if "chunk" in f.readline():
                f.close()
                for f in lFiles:
                    FastaUtils.convertFastaHeadersFromChkToChr( f,
                                                                mapChunkFile,
                                                                "%s.onChr" % ( f ) )
                    os.rename( f, "%s.onChk" % ( f ) )
                    os.rename( "%s.onChr" % ( f ), f )
            else:
                f.close()

        if multAlign == "Mafft" and not CheckerUtils.isExecutableInUserPath("mafft") :
                print "ERROR: mafft must be in your path"
                sys.exit(1)
        if multAlign == "Clustalw" and not CheckerUtils.isExecutableInUserPath("clustalw") :
                print "ERROR: clustalw must be in your path"
                sys.exit(1)
        if multAlign == "Muscle" and not CheckerUtils.isExecutableInUserPath("muscle") :
                print "ERROR: muscle must be in your path"
                sys.exit(1)       
        if multAlign == "Tcoffee" and not CheckerUtils.isExecutableInUserPath("t_coffee") :
                print "ERROR: t_coffee must be in your path"
                sys.exit(1)
        if multAlign == "Prank" and not CheckerUtils.isExecutableInUserPath("prank") :
                print "ERROR: prank must be in your path"
                sys.exit(1) 
        if multAlign in [ "Clustalw", "Muscle", "Tcoffee", "Prank" ]:
            prg = os.environ["REPET_PATH"] + "/bin/srpt%s.py" % ( multAlign )
            cmd = prg
            cmd += " -g %s_TEdenovo_%s_%s_%s" % ( projectName, smplAlign, clustHsp, multAlign )
            cmd += " -q %s" % ( os.getcwd() )
            cmd += " -Q \"%s\"" % ( config.get("build_consensus","queue") )
            if config.get("build_consensus","tmpDir" ) != "":
                cmd += " -d %s" % ( config.get("build_consensus","tmpDir") )
            if config.get("build_consensus","clean") == "yes":
                cmd += " -c"
            if multAlign == "Prank":
                cmd += " -P \"%s\"" % ( "-f=8 -noxml -notree -F -quiet" )
            elif multAlign == "Tcoffee":
                cmd += " -P \"%s\"" % ( "-type dna -output fasta_aln" )
            if os.environ[ "REPET_QUEUE" ] == "SGE":
                if not CheckerUtils.isQueueNameValid(config.get("build_consensus","queue")):
                    print "ERROR: the queue name specified in build_consensus section in your configuration file is not valid"
                    sys.exit(1)
            pL.launch( prg, cmd )
        elif multAlign in [ "Map", "MAP", "Mafft" ]:
            if multAlign in [ "Map", "MAP" ]:
                cL = MapClusterLauncher()
            elif multAlign == "Mafft":
                cL = MafftClusterLauncher()
            cL.setInputDirectory( os.getcwd() )
            cL.setQueueName( config.get("build_consensus","queue") )
            cL.setGroupIdentifier( "%s_TEdenovo_%s_%s_%s" % ( \
                    projectName, smplAlign, clustHsp, multAlign ) )
            cL.setAcronym( multAlign )
            cL.setConfigFile( "%s/%s" % ( \
                    config.get("project","project_dir"), configFileName ) )
            cL.setTemporaryDirectory( config.get("build_consensus","tmpDir" ) )
            if config.get("build_consensus","clean" ) == "yes":
                cL.setClean()
            cL.setVerbosityLevel( verbose )
            cL.setSingleProgramLauncher()
            try :
                cL.run()
            except CheckerException, e:
                print e.message
                sys.exit(1)


        # build the consensus for each MSA and concatenate them into a single fasta file
        prg = os.environ["REPET_PATH"] + "/bin/buildConsLib.py"
        cmd = prg
        cmd += " -d %s" % ( os.getcwd() )
        cmd += " -r \"*.fa_aln\""
        cmd += " -n %s" % ( config.get("build_consensus","minBasesPerSite") )
        cmd += " -o %s_%s_consensus.fa" % ( prefix1, multAlign )
        if config.get("build_consensus","clean") == "yes":
            cmd += " -c"
        cmd += " -v %i" % ( verbose - 1 )
        pL.launch( prg, cmd )

        os.chdir( ".." )

    else:
        print "ERROR: %s not implememented" % ( clustHsp )
        sys.exit(1)

    if config.get("build_consensus","clean") == "yes":
        cmd = "find . -name \"profil*\" -exec rm {} \;"
        pL.launch( "find", cmd )

    if verbose > 0:
        print "step 3 finished successfully"; sys.stdout.flush()

#------------------------------------------------------------------------------

def detectFeatures( smplAlign, clustHsp, multAlign ):

    """
    step 4: detect TE features on the consensus

    @param smplAlign: name of the alignment program (Blaster or Pals)
    @type smplAlign: string

    @param clustHsp: name of the clustering program (Grouper, Recon or Piler)
    @type clustHsp: string

    @param multAlign: name of the multiple alignment program MSA (MAP, Mafft, Muscle, Tcoffee, ClustalW or Prank)
    @type multAlign: string
    """

    if verbose > 0:
        print "beginning of step 4"
        print "self-alignment with %s" % ( smplAlign )
        print "clustering with %s" % ( clustHsp )
        print "multiple alignment with %s" % ( multAlign )
        sys.stdout.flush()

    if verbose > 0:
        print "detect features on the consensus (step 1 of TEclassifier)"; sys.stdout.flush()

    prefix = "%s_%s_%s_%s" % ( projectName, smplAlign, clustHsp, multAlign )

    if os.path.exists( "%s_TEclassif" % ( prefix ) ):
        print "ERROR: directory '%s_TEclassif/' already exists" % ( prefix )
        sys.exit(1)
    os.mkdir( "%s_TEclassif" % ( prefix ) )
    os.chdir( "%s_TEclassif" % ( prefix ) )

    if not os.path.exists( "detectFeatures" ):
        os.mkdir( "detectFeatures" )
    else:
        if verbose > 0:
            print "** Warning: directory detectFeatures/ already exists"
            sys.stdout.flush()
    os.chdir( "detectFeatures" )

    if not os.path.exists( configFileName ):
        if not os.path.exists( "../../" + configFileName ):
            print "ERROR: %s not in project directory" % ( configFileName )
            sys.exit(1)
        else:
            os.system( "ln -s ../../%s ." % ( configFileName ) )
            
    sectionName = "detect_features"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName) 
    except NoSectionError:
        print "ERROR: the section " + sectionName + " must be define in your config file : " + configFileName
        sys.exit(1)
    
    isOptionErrorRaised = False
    optionName = "TE_BLRtx"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, optionName) 
    except NoOptionError:
        print "ERROR: the option " + optionName + " must be define in " + sectionName +" in your config file : " + configFileName
        isOptionErrorRaised = True
    optionName = "TE_BLRn"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, optionName) 
    except NoOptionError:
        print "ERROR: the option " + optionName + " must be define in " + sectionName +" in your config file : " + configFileName
        isOptionErrorRaised = True
    optionName = "TE_BLRx"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, optionName) 
    except NoOptionError:
        print "ERROR: the option " + optionName + " must be define in " + sectionName +" in your config file : " + configFileName
        isOptionErrorRaised = True
    
    optionName = "HG_BLRn"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, optionName) 
    except NoOptionError:
        print "ERROR: the option " + optionName + " must be define in " + sectionName +" in your config file : " + configFileName
        isOptionErrorRaised = True
    
    optionName = "te_hmmer"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, optionName) 
    except NoOptionError:
        print "ERROR: the option " + optionName + " must be define in " + sectionName +" in your config file : " + configFileName
        isOptionErrorRaised = True
    
    if  isOptionErrorRaised : 
        sys.exit(1)
     
    if (config.get("detect_features","TE_BLRtx") == "yes" or config.get("detect_features","TE_BLRn") == "yes"):
        optionName = "TE_nucl_bank"
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, optionName) 
        except NoOptionError:
            print "ERROR: the option " + optionName + " must be define in " + sectionName +" in your config file : " + configFileName
            sys.exit(1)
        TE_nucl_bank = config.get("detect_features","TE_nucl_bank")
        if not os.path.exists( TE_nucl_bank ):
            if not os.path.exists( "../../" + TE_nucl_bank ):
                print "ERROR: %s not in project directory" % ( TE_nucl_bank )
                sys.exit(1)
            else:
                os.system( "ln -s ../../%s ." % ( TE_nucl_bank ) )

    if config.get("detect_features","TE_BLRx") == "yes":
        optionName = "TE_prot_bank"
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, optionName) 
        except NoOptionError:
            print "ERROR: the option " + optionName + " must be define in " + sectionName +" in your config file : " + configFileName
            sys.exit(1)
        TE_prot_bank = config.get("detect_features","TE_prot_bank")
        if not os.path.exists( TE_prot_bank ):
            if not os.path.exists( "../../" + TE_prot_bank ):
                print "ERROR: %s not in project directory" % ( TE_prot_bank )
                sys.exit(1)
            else:
                os.system( "ln -s ../../%s ." % ( TE_prot_bank ) )

    if config.get("detect_features","HG_BLRn") == "yes":
        optionName = "HG_nucl_bank"
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, optionName) 
        except NoOptionError:
            print "ERROR: the option " + optionName + " must be define in " + sectionName +" in your config file : " + configFileName
            sys.exit(1)
        HG_nucl_bank = config.get("detect_features","HG_nucl_bank")
        if not os.path.exists( HG_nucl_bank ):
            if not os.path.exists( "../../" + HG_nucl_bank ):
                print "ERROR: %s not in project directory" % ( HG_nucl_bank )
                sys.exit(1)
            else:
                os.system( "ln -s ../../%s ." % ( HG_nucl_bank ) )

    if config.get("detect_features","te_hmmer") == "yes":
        optionName = "te_hmm_profiles"
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, optionName) 
        except NoOptionError:
            print "ERROR: the option " + optionName + " must be define in " + sectionName +" in your config file : " + configFileName
            sys.exit(1)
        HMM_profiles = config.get("detect_features","te_hmm_profiles")
        if not os.path.exists( HMM_profiles ):
            if not os.path.exists( "../../" + HMM_profiles ):
                print "ERROR: %s not in project directory" % ( HMM_profiles )
                sys.exit(1)
            else:
                os.system( "ln -s ../../%s ." % ( HMM_profiles ) )
                
    if config.get("detect_features","rDNA_BLRn") == "yes":
        optionName = "rDNA_bank"
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, optionName) 
        except NoOptionError:
            print "ERROR: the option " + optionName + " must be define in " + sectionName +" in your config file : " + configFileName
            sys.exit(1)
        rDNA_bank = config.get("detect_features","rDNA_bank")
        if not os.path.exists( rDNA_bank ):
            if not os.path.exists( "../../" + rDNA_bank ):
                print "ERROR: %s not in project directory" % ( rDNA_bank )
                sys.exit(1)
            else:
                os.system( "ln -s ../../%s ." % ( rDNA_bank ) )


    if not os.path.exists( "%s_consensus.fa" % ( prefix ) ):

        # shorten the headers of the consensus
        prg = os.environ["REPET_PATH"] + "/bin/shortenHeaderConsensus.py"
        cmd = prg
        cmd += " -p %s" % ( projectName )
        cmd += " -s %s" % ( smplAlign )
        cmd += " -m %s" % ( multAlign )
        cmd += " -v %i" % ( verbose - 1 )
        if "Grouper" in clustHsp or "Grp" in clustHsp:
            cmd += " -c Grouper"
            cmd += " -i ../../%s_%s_Grouper_%s/%s_%s_Grouper_%s_consensus.fa" % ( projectName, smplAlign, multAlign, projectName, smplAlign, multAlign )
            cmd += " -o %s_%s_Grouper_%s_consensus_shortH.fa" % ( projectName, smplAlign, multAlign )
            pL.launch( prg, cmd )
        if "Rec" in clustHsp:
            cmd += " -c Recon"
            cmd += " -i ../../%s_%s_Recon_%s/%s_%s_Recon_%s_consensus.fa" % ( projectName, smplAlign, multAlign, projectName, smplAlign, multAlign )
            cmd += " -o %s_%s_Recon_%s_consensus_shortH.fa" % ( projectName, smplAlign, multAlign )
            pL.launch( prg, cmd )
        if "Pil" in clustHsp:
            cmd += " -c Piler"
            cmd += " -i ../../%s_%s_Piler_%s/%s_%s_Piler_%s_consensus.fa" % ( projectName, smplAlign, multAlign, projectName, smplAlign, multAlign )
            cmd += " -o %s_%s_Piler_%s_consensus_shortH.fa" % ( projectName, smplAlign, multAlign )
            pL.launch( prg, cmd )

        # concatenate all the consensus
        lFiles = glob.glob( "%s_%s_*_%s_consensus_shortH.fa" % ( projectName, smplAlign, multAlign ) )
        lFiles.sort()
        for resFile in lFiles:
            prg = "cat"
            cmd = prg
            cmd += " %s >> %s_consensus.fa" % ( resFile, prefix )
            pL.launch( prg, cmd )

        # clean
        if os.path.exists( "%s_%s_Grouper_%s_consensus_shortH.fa" % ( projectName, smplAlign, multAlign ) ):
            os.system( "rm -f %s_%s_Grouper_%s_consensus_shortH.fa" % ( projectName, smplAlign, multAlign ) )
        if os.path.exists( "%s_%s_Recon_%s_consensus_shortH.fa" % ( projectName, smplAlign, multAlign ) ):
            os.system( "rm -f %s_%s_Recon_%s_consensus_shortH.fa" % ( projectName, smplAlign, multAlign ) )
        if os.path.exists( "%s_%s_Piler_%s_consensus_shortH.fa" % ( projectName, smplAlign, multAlign ) ):
            os.system( "rm -f %s_%s_Piler_%s_consensus_shortH.fa" % ( projectName, smplAlign, multAlign ) )

    # launch TEclassifier.py step 1
    prg = os.environ["REPET_PATH"] + "/bin/TEclassifier.py"
    cmd = prg
    cmd += " -p %s" % ( prefix )
    cmd += " -c %s" % ( configFileName )
    cmd += " -i %s_consensus.fa" % ( prefix )
    cmd += " -s 1"
    cmd += " -v %i" % ( verbose - 1 )
    pL.launch( prg, cmd )

    os.chdir ( "../.." )

    if verbose > 0:
        print "step 4 finished successfully"; sys.stdout.flush()

#------------------------------------------------------------------------------

def classifConsensus( smplAlign, clustHsp, multAlign ):

    """
    step 5: classify the consensus according to their features

    @param smplAlign: name of the alignment program (Blaster or Pals)
    @type smplAlign: string

    @param clustHsp: name of the clustering program (Grouper, Recon or Piler)
    @type clustHsp: string

    @param multAlign: name of the multiple alignment program MSA (MAP, Mafft, Muscle, Tcoffee, ClustalW or Prank)
    @type multAlign: string
    """

    if verbose > 0:
        print "beginning of step 5"
        print "self-alignment with %s" % ( smplAlign )
        print "clustering with %s" % ( clustHsp )
        print "multiple alignment with %s" % ( multAlign )
        print "classify the consensus according to their features (step 2 of TEclassifier)"
        sys.stdout.flush()

    prefix = "%s_%s_%s_%s" % ( projectName, smplAlign, clustHsp, multAlign )

    if not os.path.exists( prefix + "_TEclassif" ):
        print "ERROR: run step 4 before step 5"
        help()
        sys.exit(1)
    else:
        os.chdir( prefix + "_TEclassif" )

    if os.path.exists( "./classifConsensus" ):
        print "ERROR: directory %s_TEclassif/classifConsensus already exists" % ( prefix )
        sys.exit(1)
    else:
        os.mkdir( "classifConsensus" )
    os.chdir( "classifConsensus" )

    if not os.path.exists( "%s/%s" % ( projectDir, configFileName ) ):
        print "ERROR: %s not in project directory" % ( configFileName )
        sys.exit(1)
    else:
        os.system( "ln -s %s/%s ." % ( projectDir, configFileName ) )

    os.system( "ln -s ../detectFeatures/%s_consensus.fa ." % ( prefix ) )

    # classify consensus according to their features and remove redundancy
    prg = os.environ["REPET_PATH"] + "/bin/TEclassifier.py"
    cmd = prg
    cmd += " -p %s" % ( prefix )
    cmd += " -c %s" % ( configFileName )
    cmd += " -i %s_consensus.fa" % ( prefix )
    cmd += " -s 2"
    cmd += " -v %i" % ( verbose - 1 )
    pL.launch( prg, cmd )

    # shorten the headers
    prg = os.environ["REPET_PATH"] + "/bin/shortenHeaderClassif.py"
    cmd = prg
    cmd += " -i %s_uniq_TEclassifier.fa" % ( prefix )
    cmd += " -c"
    cmd += " -s \"_\""
    cmd += " -o %s_denovoLibTEs.fa" % ( projectName )
    cmd += " -v %i" % ( verbose - 1 )
    pL.launch( prg, cmd )

    pL.launch( "mv", "mv %s_denovoLibTEs.fa ../.." % ( projectName ) )
    os.chdir ( "../.." )

    if verbose > 0:
        print "step 5 finished successfully"; sys.stdout.flush()

#------------------------------------------------------------------------------

if __name__ == "__main__":
    main()
