#!/usr/bin/env python

"""
This program removes redundancy among queries compared to subjects.
"""

import os
import sys
import getopt
import exceptions
import logging
import time
import shutil

if not os.environ.has_key( "REPET_PATH" ):
    print "ERROR: no environment variable REPET_PATH"
    sys.exit(1)
sys.path.append( os.environ["REPET_PATH"] )
import pyRepet.seq.fastaDB
import pyRepet.launcher.programLauncher


def help():
    """
    Give the list of the command-line options.
    """
    print
    print "usage:",sys.argv[0]," [ options ]"
    print "options:"
    print "     -h: this help"
    print "     -q: fasta filename of the queries"
    print "     -s: fasta filename of the subjects (same as queries if not specified)"
    print "     -o: output queries fasta filename (default=qryFileName+'.uniq')"
    print "     -i: identity threshold (default=0.95)"
    print "     -l: length threshold (default=0.98)"
    print "     -e: E-value threshold (default=1e-10)"
    print "     -w: minimum cutter word for BLAST (default=11)"
    print "     -f: length filter for Blaster (default=100)"
    print "     -Q: queue name to run in parallel"
    print "     -d: absolute path to the temporary directory"
    print "     -C: name of the configuration file (compulsory with '-Q')"
    print "     -n: max. number of jobs (default=100,given a min. of 1 query per job)"
    print "     -c: clean"
    print "     -v: verbose (default=0/1/2)"
    print


def main():
    """
    This program removes redundancy among queries compared to subjects.
    """
    qryFileName = ""
    sbjFileName = ""
    outFileName = ""
    thresIdentity = 0.95   # remove the seq if it is identical to 95% of another seq
    thresLength = 0.98   # and if its length is 98% of that seq
    threshEvalue = "1e-10"
    wordBlast = 11
    lengthFilter = 100
    useCluster = False
    queue = ""
    tmpDir = ""
    configFileName = ""
    maxNbJobs = 100
    clean = False
    verbose = 0
    try:
        opts, args = getopt.getopt(sys.argv[1:],"hq:s:o:L:i:l:e:w:f:Q:d:C:n:cv:")
    except getopt.GetoptError, err:
        print str(err)
        help()
        sys.exit(1)
    for o,a in opts:
        if o == "-h":
            help()
            sys.exit(0)
        elif o == "-q":
            qryFileName = a 
        elif o == "-s":
            sbjFileName = a 
        elif o == "-o":
            outFileName = a
        elif o == "-L":
            logFileName = a
        elif o == "-i":
            thresIdentity = float(a) 
        elif o == "-l":
            thresLength = float(a)
        elif o == "-e":
            threshEvalue = a
        elif o == "-w":
            wordBlast = int(a)
        elif o == "-f":
            lengthFilter = int(a)
        elif o == "-Q":
            useCluster = True
            queue = a
        elif o == "-d":
            tmpDir = a
        elif o == "-C":
            configFileName = a
        elif o == "-n":
            maxNbJobs = int(a)
        elif o == "-c":
            clean = True
        elif o == "-v":
            verbose = int(a)
    if thresIdentity > 1.0 or thresLength > 1.0:
        print "ERROR: thresholds must be <= 1.0"
        sys.exit(1)
    if qryFileName == "":
        print "ERROR: missing input file"
        help()
        sys.exit(1)
    if useCluster == True:
        if configFileName == "":
            print "ERROR: missing configuration file"
            sys.exit(1)
        if not os.path.exists( configFileName ):
            print "ERROR: configuration file doesn't exist"
            sys.exit(1)

    #--------------------------------------------------------------------------

    if verbose > 0:
        print "START %s (%s)" % ( sys.argv[0].split("/")[-1],
                                  time.strftime("%m/%d/%Y %H:%M:%S") )
        sys.stdout.flush()

    if outFileName == "":
        outFileName = "%s.uniq" % ( qryFileName )
        
    uniqId = "%s-%s" % ( time.strftime("%Y%m%d%H%M%S"), os.getpid() )

    # create the 'log' file
    logFileName = "%s_rmvRedundancy_%s.log" % ( outFileName, uniqId )
    handler = logging.FileHandler( logFileName )
    formatter = logging.Formatter( "%(asctime)s %(levelname)s: %(message)s" )
    handler.setFormatter( formatter )
    logging.getLogger('').addHandler( handler )
    logging.getLogger('').setLevel( logging.DEBUG )
    logging.info( "started" )

    if not os.path.exists( qryFileName ):
        string = "query file '%s' doesn't exist" % ( qryFileName )
        logging.error( string )
        print "ERROR: %s" % ( string )
        sys.exit(1)
    if sbjFileName != "":
        if not os.path.exists( sbjFileName ):
            string = "subject file '%s' doesn't exist" % ( sbjFileName )
            logging.error( string )
            print "ERROR: %s" % ( string )
            sys.exit(1)
    else:
        sbjFileName = qryFileName

    logging.info( "remove redundancy among '%s' (queries) compare to '%s' (subjects)" % ( qryFileName, sbjFileName ) )

    #--------------------------------------------------------------------------

    # check the input files are not empty, otherwise exit

    nbSeqQry = pyRepet.seq.fastaDB.dbSize( qryFileName )
    if nbSeqQry == 0:
        string = "query file is empty"
        logging.info( string )
        print "WARNING: %s" % ( string )
        logging.info( "finished" )
        sys.exit(0)

    nbSeqSbj = pyRepet.seq.fastaDB.dbSize( sbjFileName )
    if sbjFileName != qryFileName:
        nbSeqSbj = pyRepet.seq.fastaDB.dbSize( sbjFileName )
        if nbSeqSbj == 0:
            string = "subject file is empty"
            logging.info( string )
            print "WARNING: %s" % ( string )
            logging.info( "finished" )
            sys.exit(0)

    #--------------------------------------------------------------------------

    pL = pyRepet.launcher.programLauncher.programLauncher()

    # work in a temporary directory
    tmpDir = "tmp%s" % ( uniqId )
    if os.path.exists( tmpDir ):
        shutil.rmtree( tmpDir )
    os.mkdir( tmpDir )
    os.chdir( tmpDir )
    os.system( "ln -s ../%s ." % ( qryFileName ) )
    if sbjFileName != qryFileName:
        os.system( "ln -s ../%s ." % ( sbjFileName ) )
    os.system( "ln -s ../%s ." % ( logFileName ) )

    # shorten sequence headers
    if sbjFileName == qryFileName:
        # sort sequences by increasing length
        os.system( "mv %s %s.initOrder" % ( qryFileName, qryFileName ) )
        pyRepet.seq.fastaDB.sortSequencesByIncreasingLength( "%s.initOrder" % ( qryFileName ),
                                                             qryFileName,
                                                             verbose-1 )
        pyRepet.seq.fastaDB.shortenSeqHeaders( inFileName=qryFileName )
    else:
        pyRepet.seq.fastaDB.shortenSeqHeaders( inFileName=qryFileName, pattern="qry" )
        pyRepet.seq.fastaDB.shortenSeqHeaders( inFileName=sbjFileName, pattern="sbj" )

    #--------------------------------------------------------------------------

    # case not in parallel
    if useCluster == False:

        # run Blaster
        prg = os.environ["REPET_PATH"] + "/bin/blaster"
        cmd = prg
        cmd += " -q %s.shortH" % ( qryFileName )
        cmd += " -s %s.shortH" % ( sbjFileName )
        cmd += " -a"
        cmd += " -I %i" % ( int(100*thresIdentity) )
        cmd += " -L %i" % ( lengthFilter )
        cmd += " -E %s" % ( threshEvalue )
        cmd += " -B %s.shortH_vs_%s.shortH" % ( qryFileName, sbjFileName )
        pL.launch( prg, cmd )

        # run Matcher
        prg = os.environ["REPET_PATH"] + "/bin/matcher"
        cmd = prg
        cmd += " -m %s.shortH_vs_%s.shortH.align" % ( qryFileName, sbjFileName )
        cmd += " -q %s.shortH" % ( qryFileName )
        cmd += " -s %s.shortH" % ( sbjFileName )
        cmd += " -j"
        cmd += " -a"
        pL.launch( prg, cmd )

    #--------------------------------------------------------------------------

    # case in parallel
    elif useCluster == True:

        os.system( "ln -s ../%s ." % ( configFileName ) )

        # run Blaster + Matcher
        prg = os.environ["REPET_PATH"] + "/bin/launchBlasterMatcherPerQuery.py"
        cmd = prg
        cmd += " -q %s.shortH" % ( qryFileName )
        cmd += " -s %s.shortH" % ( sbjFileName )
        cmd += " -Q %s" % ( queue )
        cmd += " -d %s" % ( os.getcwd() )
        cmd += " -C %s" % ( configFileName )
        cmd += " -n %i" % ( maxNbJobs )
        cmd += " -B \"-a -I %s -L %s -E %s\"" % ( int(100*thresIdentity), lengthFilter, threshEvalue )
        cmd += " -M \"%s\"" % ( "-j -a" )
        cmd += " -Z tab"
        if clean == True:
            cmd += " -c"
        cmd += " -v %i" % ( verbose - 1 )
        pL.launch( prg, cmd )

    #--------------------------------------------------------------------------

    # filter the resulting 'tab' file
    prg = os.environ["REPET_PATH"] + "/bin/filterOutMatcher.py"
    cmd = prg
    cmd += " -q %s.shortH" % ( qryFileName )
    if sbjFileName != qryFileName:
        cmd += " -s %s.shortH" % ( sbjFileName )
    cmd += " -m %s.shortH_vs_%s.shortH.align.match.tab" % ( qryFileName, sbjFileName )
    cmd += " -o %s.shortH.filtered" % ( qryFileName )
    cmd += " -i %f" % ( thresIdentity )
    cmd += " -l %f" % ( thresLength )
    cmd += " -L %s" % ( logFileName )
    cmd += " -v %i" % ( verbose )
    pL.launch( prg, cmd )

    # retrieve initial headers
    pyRepet.seq.fastaDB.retrieveInitSeqHeaders( "%s.shortH.filtered" % ( qryFileName ),
                                                "%s.shortHlink" % ( qryFileName ),
                                                outFileName )

    # future improvement: give file '.shortHlink' to 'filterOutMatcher.py' so that it saves the information about which match removed which query, with the true headers

    os.system( "mv %s .." % ( outFileName ) )
    os.chdir( ".." )
    if clean == True:
        shutil.rmtree( tmpDir )

    logging.info( "finished" )

    if verbose > 0:
        print "END %s (%s)" % ( sys.argv[0].split("/")[-1],
                                time.strftime("%m/%d/%Y %H:%M:%S") )
        sys.stdout.flush()

    return 0


if __name__ == "__main__":
    main()
