#!/usr/bin/env python

import os
import sys
import getopt

from pyRepetUnit.commons.coord.Align import Align
from pyRepet.coord.Match import Match


class ChangeSequenceHeaders( object ):
    
    def __init__( self ):
        self._name = "ChangeSequenceHeaders"
        self._inFile = ""
        self._format = ""
        self._step = 0
        self._prefix = "seq"
        self._linkFile = ""
        self._outFile = ""
        self._verbose = 0
        
        
    def help( self ):
        print
        print "usage: %s.py [ options ]" % ( self._name )
        print "options:"
        print "     -h: this help"
        print "     -i: name of the input file"
        print "     -f: format of the input file (fasta/newick/align/tab)"
        print "     -s: step (1: shorten headers / 2: retrieve initial headers)"
        print "     -p: prefix of new headers (with '-s 1', default='seq')"
        print "     -l: name of the 'link' file (with '-s 2', format=map)"
        print "     -o: name of the output file (default=inFile+'.newH'/'.initH')"
        print
        
        
    def setAttributesFromCmdLine( self ):
        try:
            opts, args = getopt.getopt(sys.argv[1:],"hi:f:s:p:l:o:v:")
        except getopt.GetoptError, err:
            sys.stderr.write( "%s\n" % ( str(err) ) )
            self.help(); sys.exit(1)
        for o,a in opts:
            if o == "-h":
                self.help(); sys.exit(0)
            elif o == "-i":
                self.setInputFile( a )
            elif o == "-f":
                self.setFormat( a )
            elif o == "-s":
                self.setStep( a )
            elif o == "-p":
                self.setPrefix( a )
            elif o == "-l":
                self.setLinkFile( a )
            elif o == "-o":
                self.setOutputFile( a )
            elif o == "-v":
                self.setVerbosityLevel( a )
                
                
    def setInputFile( self, inFile ):
        self._inFile = inFile
        
    def setFormat( self, format ):
        self._format = format
        
    def setPrefix( self, prefix ):
        self._prefix = prefix
        
    def setStep( self, step ):
        self._step = int(step)
        
    def setLinkFile( self, linkFile ):
        self._linkFile = linkFile
        
    def setOutputFile( self, outFile ):
        self._outFile = outFile
        
    def setVerbosityLevel( self, verbose ):
        self._verbose = int(verbose)
        
        
    def checkAttributes( self ):
        if self._inFile == "":
            sys.stderr.write( "ERROR: missing input file name (-i)\n" )
            self.help(); sys.exit(1)
        if not os.path.exists( self._inFile ):
            sys.stderr.write( "ERROR: input file doesn't exist (-i)\n" )
            self.help(); sys.exit(1)
        if self._format not in ["fasta","newick","align","tab"]:
            sys.stderr.write( "ERROR: unrecognized format '%s' (-f)\n" )
            self.help(); sys.exit(1)
        if self._step not in [1,2]:
            sys.stderr.write( "ERROR: missing step (-s)\n" )
            self.help(); sys.exit(1)
        if self._step == 1 and self._prefix == "":
            sys.stderr.write( "ERROR: missing prefix (-p)\n" )
            self.help(); sys.exit(1)
        if self._step == 2:
            if self._linkFile == "":
                sys.stderr.write( "ERROR: missing link file name (-l)\n" )
                self.help(); sys.exit(1)
            if not os.path.exists( self._linkFile ):
                sys.stderr.write( "ERROR: link file doesn't exist (-l)\n" )
                self.help(); sys.exit(1)
        if self._outFile == "":
            if self._step == 1:
                self._outFile = "%s.newH" % ( self._inFile )
            elif self._step == 2:
                self._outFile = "%s.initH" % ( self._inFile )
                
                
    def shortenSequenceHeadersForFastaFile( self ):
        if self._verbose > 0:
            print "shorten sequence headers for fasta file..."
            sys.stdout.flush()
        inFileHandler = open( self._inFile, "r" )
        linkFileHandler = open( self._linkFile, "w" )
        outFileHandler = open( self._outFile, "w" )
        countSeq = 0
        lengthSeq = 0
        while True:
            line = inFileHandler.readline()
            if line == "":
                break
            if line[0] == ">":
                countSeq += 1
                newHeader = "%s%i" % ( self._prefix, countSeq )
                outFileHandler.write( ">%s\n" % ( newHeader ) )
                if lengthSeq != 0:
                    linkFileHandler.write( "\t%i\t%i\n" % ( 1, lengthSeq ) )
                    lengthSeq = 0
                linkFileHandler.write( "%s\t%s" % ( newHeader, line[1:-1] ) )
            else:
                lengthSeq += len( line.replace("\n","") )
                outFileHandler.write( line )
        linkFileHandler.write( "\t%i\t%i\n" % ( 1, lengthSeq ) )
        inFileHandler.close()
        linkFileHandler.close()
        outFileHandler.close()
        if self._verbose > 0:
            print "nb of sequences: %i" % ( countSeq )
            
            
    def getLinksNewToInitialHeaders( self ):
        if self._verbose > 0:
            print "retrieve the links new->initial headers"
            sys.stdout.flush()
        dNew2Init = {}
        linkFileHandler = open( self._linkFile,"r" )
        while True:
            line = linkFileHandler.readline()
            if line == "":
                break
            tokens = line.split("\t")
            if len(tokens) == 4:
                dNew2Init[ tokens[0] ] = tokens[1]
            elif len(tokens) == 2:
                dNew2Init[ tokens[0] ] = tokens[1].split("\n")[0]
            else:
                sys.stderr.write( "ERROR: link file is badly formatted\n" )
                sys.exit(1)
        linkFileHandler.close()
        if self._verbose > 0: print "nb of links: %i" % ( len(dNew2Init.keys()) ); sys.stdout.flush()
        return dNew2Init
    
    
    def retrieveInitialSequenceHeadersForFastaFile( self, dNew2Init ):
        if self._verbose > 0:
            print "retrieve initial headers for fasta file"
            sys.stdout.flush()
        inFileHandler = open( self._inFile, "r" )
        outFileHandler = open( self._outFile, "w" )
        countSeq = 0
        while True:
            line = inFileHandler.readline()
            if line == "":
                break
            if line[0] == ">":
                countSeq += 1
                newHeader = dNew2Init[ line[1:-1] ]
                outFileHandler.write( ">%s\n" % ( newHeader ) )
            else:
                outFileHandler.write( line )
        inFileHandler.close()
        outFileHandler.close()
        if self._verbose > 0:
            print "nb of sequences: %i" % ( countSeq )
            
            
    def retrieveInitialSequenceHeadersForNewickFile( self, dNew2Init ):
        inF = open( self._inFile, "r" )
        lines = inF.readlines()
        inF.close()
        line = "".join(lines)   #.replace(";","").replace("\n","")
        outF = open( self._outFile, "w" )
        for newH in dNew2Init.keys():
            line = line.replace( newH+":", dNew2Init[newH].replace(" ","_").replace("::","-").replace(":","-").replace(",","-")+":" )
        outF.write( line )
        outF.close()
        
        
    def retrieveInitialSequenceHeadersForAlignFile( self, dNew2Init ):
        inFileHandler = open( self._inFile, "r" )
        outFileHandler = open( self._outFile, "w" )
        a = Align()
        while True:
            line = inFileHandler.readline()
            if line == "":
                break
            a.setFromTuple( line.split("\t") )
            nameToBeReplaced = a.range_query.seqname
            if dNew2Init.has_key( nameToBeReplaced ):
                a.range_query.seqname = dNew2Init[ nameToBeReplaced ]
            nameToBeReplaced = a.range_subject.seqname
            if dNew2Init.has_key( nameToBeReplaced ):
                a.range_subject.seqname = dNew2Init[ nameToBeReplaced ]
            a.write( outFileHandler )
        inFileHandler.close()
        outFileHandler.close()
        
        
    def retrieveInitialSequenceHeadersForMatchFile( self, dNew2Init ):
        inFileHandler = open( self._inFile, "r" )
        outFileHandler = open( self._outFile, "w" )
        m = Match()
        line = inFileHandler.readline()
        while True:
            line = inFileHandler.readline()
            if line == "":
                break
            m.setFromTuple( line.split("\t") )
            nameToBeReplaced = m.range_query.seqname
            if dNew2Init.has_key( nameToBeReplaced ):
                m.range_query.seqname = dNew2Init[ nameToBeReplaced ]
            nameToBeReplaced = m.range_subject.seqname
            if dNew2Init.has_key( nameToBeReplaced ):
                m.range_subject.seqname = dNew2Init[ nameToBeReplaced ]
            m.write( outFileHandler )
        inFileHandler.close()
        outFileHandler.close()
        
        
    def run( self ):
        self.checkAttributes()
        if self._step == 1:
            if self._linkFile == "":
                self._linkFile = "%s.newHlink" % ( self._inFile )
            if self._format == "fasta":
                self.shortenSequenceHeadersForFastaFile()
        if self._step == 2:
            dNew2Init = self.getLinksNewToInitialHeaders()
            if self._format == "fasta":
                self.retrieveInitialSequenceHeadersForFastaFile( dNew2Init )
            elif self._format == "newick":
                self.retrieveInitialSequenceHeadersForNewickFile( dNew2Init )
            elif self._format == "align":
                self.retrieveInitialSequenceHeadersForAlignFile( dNew2Init )
            elif self._format in [ "tab", "match" ]:
                self.retrieveInitialSequenceHeadersForMatchFile( dNew2Init )
                
                
if __name__ == "__main__":
    i = ChangeSequenceHeaders()
    i.setAttributesFromCmdLine()
    i.run()
